from .accumulator import Accumulator
from .curve import MIN_STEP
from intervalpy import Interval


class AccumulatorMap(Accumulator):

    def __init__(
            self,
            func,
            tfm,
            degree,
            is_period=True,
            interpolation=None,
            uniform=True,
            min_step=MIN_STEP):
        super().__init__(
            func,
            self._accumulator_map_scan,
            interpolation=interpolation,
            uniform=uniform,
            min_step=min_step
        )

        tfm_ags = type(self).count_positional_args(tfm)
        if tfm_ags == 0 or tfm_ags > 2:
            raise Exception('Unable to adapt function')
        self._accumulator_map_tfm_with_x = tfm_ags > 1
        self.accumulator_map_tfm = tfm

        self.degree = None
        self.period = None
        if is_period:
            self.period = degree
            if self.period is None or self.period <= 0:
                raise Exception('Min period must be a positive number')
        else:
            self.degree = int(degree)
            if self.degree <= 0:
                raise Exception('Min degree must be a positive integer')

    def _accumulator_map_scan(self, x, y, _):
        points = [[x, y]]
        if self.degree is not None:
            if self.degree != 1:
                points += self.curve.sample_points_from_x(
                    x,
                    self.degree - 1,
                    backward=True,
                    open=True,
                    min_step=self.min_step
                )
        elif self.period is not None:
            points += reversed(self.curve.sample_points(
                domain=Interval.closed_open(x - self.period, x),
                min_step=self.min_step
            ))
        else:
            raise Exception('Bad config')

        ys = [p[1] for p in points]
        return self._accumulator_map(x, ys)

    def _accumulator_map(self, x, ys):
        if self._accumulator_map_tfm_with_x:
            return self.accumulator_map_tfm(x, ys)
        else:
            return self.accumulator_map_tfm(ys)



import warnings
from .extension import Extension
from ..constant import Constant
from ..empty import Empty
from ..curve import Curve

class HarmonicExtension(Extension):

    """
    Extends an end of a function with a series of sine wave using its edge tangents.
    """

    name = "harmonic"

    def __init__(self, func, period=None, degree=3, regression_degree=5, **kwargs):
        self.period = float(period)
        self.degree = int(degree)
        self.regression_degree = int(regression_degree)
        self.regression = 0.1
        assert self.period > 0
        assert self.degree > 1
        assert self.regression_degree > 1
        self._fast_sma = Empty()
        self._extension_offset = Constant(0)
        self.smas = []
        self.macs = []
        super().__init__(func, **kwargs)
        if self.start:
            raise Exception('Weave extension currently only supports extending the end')

    def update_extension(self):
        x = self.curve.domain.end
        y = self._fast_sma.y(x)
        y_ext = self.end_func.y(x)
        
        if y_ext is None:
            return

        y_offset = y - y_ext
        if y_offset != 0:
            self._extension_offset.value = self._extension_offset.value + y_offset

    def create_extension_func(self, start=False):
        base_line = Constant(0)
        hsmas = reversed(self.curve.harmonic_smas(
            self.period,
            self.degree,
            stride=1,
            is_period=True
        ))
        sma_tangents = []
        smas = []
        macs = []
        mac_period_coef = 2.5

        for hsma in hsmas:
            sma = hsma.offset(hsma.period * -0.5)
            sma_tangent = sma.extension(
                'tangent',
                regression_period=hsma.period * self.regression,
                start=False,
                end=True
            )
            if not sma_tangent.domain.is_positive_infinite:
                # Not enough data for this period
                warnings.warn('Not enough data to extend with all degrees of harmonic')
                continue
            sma_tangents.append(sma_tangent)
            if len(sma_tangents) == 1:
                base_line = sma_tangent
                smas.append(sma_tangent)
            else:
                mac_period = hsma.period * mac_period_coef
                ref_sma = smas[-1]
                mac = (sma - ref_sma).extension(
                    'sin',
                    period=mac_period,
                    regression_period=hsma.period * self.regression,
                    start=False,
                    end=True
                )
                macs.append(mac)

                sma_ext = type(self).first([sma, ref_sma + mac])
                smas.append(sma_ext)

        # Save extension
        self._fast_sma = sma_tangents[-2] + macs[-1]
        self.macs = macs
        self.smas = smas

        return Curve.add_many([self._extension_offset, base_line] + macs)



import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Constant(Curve):

    @property
    def value(self):
        return self._value

    @classmethod
    def zero(cls):
        return _zero

    @value.setter
    def value(self, value):
        if value == self._value:
            return
        self.begin_update(self.domain)
        self._value = value
        self.end_update(self.domain)

    def get_domain(self):
        return self._const_interval

    def __init__(self, y, domain=None):
        super().__init__(min_step=math.inf)
        self._value = y
        if domain is not None:
            domain = Interval.parse(domain)
        else:
            domain = Interval.infinite()
        self._const_interval = domain

    def __repr__(self):
        try:
            return str(self.value)
        except Exception as e:
            return super().__repr__() + f'({e})'

    def y(self, x):
        return self.value

    def d_y(self, x, forward=False, min_step=MIN_STEP, limit=None):
        return 0.0

    def x(self, y):
        if y == self.value:
            return 0.0
        return None

_zero = Constant(0)



import math
import weakref
import inspect
import arrow
import numpy as np
from numbers import Number
from collections.abc import Sequence, Mapping
from intervalpy import Interval
from .const import GOLD
from . import util

MIN_STEP = 1e-5

# TODO: Implement Duration and use its next ad previous methods
# Or make a super class which is not tied to a time interval.

_func_obj = None

class Curve:

    _token_counter = 0

    @classmethod
    def empty(cls):
        from .empty import Empty
        return Empty()

    @property
    def min_step(self):
        return self._min_step

    @min_step.setter
    def min_step(self, value):
        self._min_step = value

    @property
    def domain(self):
        if self.needs_domain_update:
            self._domain = self.get_domain()
        return self._domain

    @property
    def update_interval(self):
        return self._begin_update_interval

    @property
    def is_updating(self):
        return not self.update_interval.is_empty

    def __init__(self, min_step=None):
        self.name = None
        self._domain = None
        self._observer_data = {}
        self._ordered_observer_tokens = []
        self._begin_update_interval = Interval.empty()
        self._end_update_interval = Interval.empty()
        self.min_step = min_step

    def __call__(self, *args):
        return self.y(args[0])

    def __repr__(self):
        try:
            if bool(self.name):
                return self.name
            return f'{type(self).__name__}("{self.domain}")'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def y(self, x):
        raise Exception("Not implemented")

    def y_start(self):
        return self.y(self.domain.start)

    def y_end(self):
        return self.y(self.domain.end)

    def d_y(self, x, forward=False, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        if forward:
            x1 = self.x_next(x, min_step=min_step, limit=limit)
        else:
            x1 = self.x_previous(x, min_step=min_step, limit=limit)

        if x1 is None:
            return None
        y1 = self.y(x1)
        if y1 is None:
            return None
        y = self.y(x)
        if y is None:
            return None
        if x1 == x:
            dy = math.inf if y1 >= y else -math.inf
            if not forward:
                dy = -dy
        else:
            dy = (y1 - y) / (x1 - x)
        return dy

    def x(self, y):
        raise Exception("Not implemented")

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        if math.isinf(min_step):
            x1 = self.domain.end
        else:
            x1 = x + min_step
        if limit is not None and x1 > limit:
            x1 = limit
        if not self.domain.contains(x1, enforce_start=False):
            return None
        return x1

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        if math.isinf(min_step):
            x1 = self.domain.start
        else:
            x1 = x - min_step
        if limit is not None and x1 < limit:
            x1 = limit
        if not self.domain.contains(x1, enforce_end=False):
            return None
        return x1

    def previous_point(self, x, min_step=MIN_STEP):
        x1 = self.x_previous(x, min_step=min_step)
        if x1 is None:
            return None
        y1 = self.y(x1)
        return (x1, y1)

    def next_point(self, x, min_step=MIN_STEP):
        x1 = self.x_next(x, min_step=min_step)
        if x1 is None:
            return None
        y1 = self.y(x1)
        return (x1, y1)

    def get_domain(self):
        return Interval.empty()

    def resolve_min_step(self, min_step):
        if min_step is None and self.min_step is None:
            return None
        elif min_step is None:
            return self.min_step
        elif self.min_step is None:
            return min_step
        else:
            return max(min_step, self.min_step)

    def sample_points(self, domain=None, min_step=MIN_STEP, step=None):
        min_step = self.resolve_min_step(min_step)
        if domain is None:
            domain = self.domain
        else:
            domain = Interval.intersection([self.domain, domain])
        if domain.is_empty:
            return []
        elif not domain.is_finite:
            raise Exception("Cannot sample points on an infinite domain {}. Specify a finite domain.".format(domain))
        x_start, x_end = domain
        x_end_bin = round(x_end / min_step) if min_step is not None else x_end
        if domain.start_open:
            points = []
        else:
            points = [(x_start, self.y(x_start))]
        if step is not None:
            x = x_start + step
            while x <= x_end:
                y = self.y(x)
                points.append((x, y))
                x += step
        elif min_step is not None and min_step > 0:
            x = self.x_next(x_start, min_step=min_step, limit=x_end)
            while x is not None and x <= x_end:
                y = self.y(x)
                points.append((x, y))
                x_bin = round(x / min_step) if min_step is not None else x
                if x_bin == x_end_bin:
                    break
                x1 = self.x_next(x, min_step=min_step, limit=x_end)
                if x1 is not None:
                    x1_bin = round(x1 / min_step) if min_step is not None else x1
                    if x1_bin <= x_bin:
                        raise Exception('Next x value {} should be greater than the previous x value {} by at least the minimum step of {}'.format(x1, x, min_step))
                x = x1

            if not domain.end_open and points[-1][0] != x_end:
                points.append((x_end, self.y(x_end)))
        else:
            raise Exception("Bad functions sample parameters.")

        return points

    def sample_points_from_x(self, x, limit, backward=False, open=False, min_step=None):
        assert limit is not None
        if limit < 0:
            limit = -limit
            backward = not backward
        min_step = self.resolve_min_step(min_step)
        points = []
        x1 = x
        i = 0

        if not open:
            if x is None:
                return points
            y = self.y(x)
            if y is None:
                return points
            i += 1

        while limit is None or i < limit:
            if not backward:
                x1 = self.x_next(x1, min_step=min_step)
            else:
                x1 = self.x_previous(x1, min_step=min_step)
            if x1 is None:
                break
            y1 = self.y(x1)
            if y1 is None:
                break
            points.append((x1, y1))
            i += 1

        return points

    def get_range(self, domain=None, **kwargs):
        points = self.sample_points(domain=domain, **kwargs)
        low = None
        high = None
        for p in points:
            if low is None or p[1] < low:
                low = p[1]
            if high is None or p[1] > high:
                high = p[1]
        if low is None or high is None:
            return Interval.empty()
        return Interval(low, high)

    def minimise(self, x, min_step=MIN_STEP, step=None, max_iterations=1000):
        x_min = x
        x_min_previous = None
        iterations = 0
        while iterations < max_iterations:
            iterations += 1
            y = self.y(x_min)
            if y is None:
                return x_min_previous
            dy0 = self.d_y(x_min, forward=False)
            dy1 = self.d_y(x_min, forward=True)
            forward = True
            if dy0 is None and dy1 is None:
                return x_min
            elif dy0 is None:
                if dy1 <= 0:
                    forward = True
                else:
                    # Sloping into null value
                    return None
            elif dy1 is None:
                if dy0 >= 0:
                    forward = False
                else:
                    # Sloping into null value
                    return None
            else:
                if dy0 * dy1 < 0 and dy0 <= 0 and dy1 >= 0:
                    # Found minimum
                    return x_min

                if dy0 * dy1 < 0:
                    # Found maximum
                    forward = abs(dy0) < abs(dy1)
                else:
                    # On slope
                    forward = dy1 < 0

            x_min_previous = x_min
            if forward:
                if step is not None:
                    x_min += step
                else:
                    x_min = self.x_next(x_min, min_step=min_step)
            else:
                if step is not None:
                    x_min -= step
                else:
                    x_min = self.x_previous(x_min, min_step=min_step)
        return x_min

    def maximise(self, x, min_step=MIN_STEP, step=None, max_iterations=1000):
        x_max = x
        x_max_previous = None
        iterations = 0
        while iterations < max_iterations:
            iterations += 1
            y = self.y(x_max)
            if y is None:
                return x_max_previous
            dy0 = self.d_y(x_max, forward=False)
            dy1 = self.d_y(x_max, forward=True)
            forward = True
            if dy0 is None and dy1 is None:
                return x_max
            elif dy0 is None:
                if dy1 >= 0:
                    forward = True
                else:
                    # Sloping into null value
                    return None
            elif dy1 is None:
                if dy0 <= 0:
                    forward = False
                else:
                    # Sloping into null value
                    return None
            else:
                if dy0 * dy1 < 0 and dy0 >= 0 and dy1 <= 0:
                    # Found maximum
                    return x_max

                if dy0 * dy1 < 0:
                    # Found minimum
                    forward = abs(dy0) < abs(dy1)
                else:
                    # On slope
                    forward = dy1 > 0

            x_max_previous = x_max
            if forward:
                if step is not None:
                    x_max += step
                else:
                    x_max = self.x_next(x_max, min_step=min_step)
            else:
                if step is not None:
                    x_max -= step
                else:
                    x_max = self.x_previous(x_max, min_step=min_step)
        return x_max

    def regression(self, domain=None, min_step=MIN_STEP, step=None):
        points = self.sample_points(domain=domain, min_step=min_step, step=step)
        for p in points:
            if p[1] is None:
                return None
        count = len(points)
        if count < 2:
            return None
            
        from .line import Line
        if count == 2:
            return Line(p1=points[0], p2=points[1])
        xy = np.vstack(points)
        x = xy[:,0]
        y = xy[:,1]
        A = np.array([x, np.ones(count)])

        # Regression
        w = np.linalg.lstsq(A.T, y, rcond=None)[0]
        m = w[0]
        c = w[1]

        return Line(const=c, slope=m)

    def add_observer(self, *obj, domain=None, begin=None, end=None, autoremove=False, prioritize=False):
        if begin is None and end is None:
            return 0
        
        Curve._token_counter += 1
        token = Curve._token_counter
        domain = Interval.parse(domain, default_inf=True)
        obj_ref = None

        if len(obj) != 0:
            if autoremove:
                # Remove observer automatically
                obj_ref = weakref.ref(obj[0], lambda _: self.remove_observer(token))
            else:
                # Calling remove_observer() is required
                obj_ref = weakref.ref(obj[0])
        elif autoremove:
            raise Exception('Autoremoving an observer requires an object')

        # Do the callback functions require the domain?
        begin_with_interval = False
        end_with_interval = False
        if begin:
            begin_with_interval = util.count_positional_args(begin) == 1
        if end:
            end_with_interval = util.count_positional_args(end) == 1

        # TODO: does saving strong references to callbacks create a retain cycle?
        self._observer_data[token] = (obj_ref, domain, begin, end, begin_with_interval, end_with_interval)
        if prioritize:
            self._ordered_observer_tokens.insert(0, token)
        else:
            self._ordered_observer_tokens.append(token)
        return token

    def remove_observer(self, token_or_obj):
        if isinstance(token_or_obj, Number):
            if token_or_obj in self._observer_data:
                del self._observer_data[token_or_obj]
                self._ordered_observer_tokens.remove(token_or_obj)
        else:
            for token in list(self._ordered_observer_tokens):
                obj_ref = self._observer_data[token][0]
                if obj_ref is not None:
                    obj = obj_ref()
                    if obj is None or obj == token_or_obj:
                        del self._observer_data[token]
                        self._ordered_observer_tokens.remove(token)

    def begin_update(self, domain):
        if domain.is_empty or self._begin_update_interval.is_superset_of(domain):
            return
        self._begin_update_interval = Interval.union([self._begin_update_interval, domain])
        for token in self._ordered_observer_tokens:
            _, callback_interval, callback, _, callback_with_interval, _ = self._observer_data[token]
            if callback_interval is None or domain.intersects(callback_interval):
                if callback is not None:
                    if callback_with_interval:
                        callback(domain)
                    else:
                        callback()

    def end_update(self, domain):
        if domain.is_empty or self._end_update_interval.is_superset_of(domain):
            return
        self._end_update_interval = Interval.union([self._end_update_interval, domain])
        if not self._end_update_interval.is_superset_of(self._begin_update_interval):
            # Keep collecting updates
            return
        
        # Updates complete
        update_interval = self._end_update_interval
        self._begin_update_interval = Interval.empty()
        self._end_update_interval = Interval.empty()
        self.set_needs_interval_update()
        for token in list(self._ordered_observer_tokens):
            _, callback_interval, _, callback, _, callback_with_interval = self._observer_data[token]
            if callback_interval is None or update_interval.intersects(callback_interval):
                if callback is not None:
                    if callback_with_interval:
                        callback(update_interval)
                    else:
                        callback()

    @property
    def needs_domain_update(self):
        return self._domain is None

    def set_needs_interval_update(self):
        self._domain = None

    def map(self, tfm, skip_none=False, name=None, **kwargs):
        from .map import Map
        return Map(self, tfm, skip_none=skip_none, name=name, **kwargs)

    def accumulator_map(self, tfm, degree, is_period=False, interpolation=None, min_step=MIN_STEP, uniform=True):
        from .accumulator_map import AccumulatorMap
        return AccumulatorMap(
            self,
            tfm,
            degree,
            is_period=is_period,
            interpolation=interpolation,
            min_step=min_step,
            uniform=uniform
        )

    def offset(self, x, duration=None):
        from .offset import Offset
        return Offset(self, x, duration=duration)

    def add(self, func):
        return Curve.add_many([self, func])

    def subtract(self, func):
        return Curve.subtract_many([self, func])

    def multiply(self, func):
        return Curve.multiply_many([self, func])

    def divide(self, func):
        return Curve.divide_many([self, func])

    def pow(self, power):
        return type(self).pow_many([self, power])

    def raised(self, base):
        return type(self).pow_many([base, self])

    def log(self, base=math.e):
        return type(self).log_many([self, base])

    def integral(self, const=0, interpolation=None, uniform=True):
        from .integral import Integral

        return Integral(self, const=const, interpolation=interpolation, uniform=uniform)

    def additive_inverse(self):
        return self.map(_additive_inverse)

    def multiplicative_inverse(self):
        return self.map(_multiplicative_inverse)

    def abs(self):
        return self.map(_abs)

    def blend(self, func, x_blend_start, x_blend_stop):
        from .aggregate import Aggregate
        from .piecewise import Piecewise
        
        x_blend_period = x_blend_stop - x_blend_start
        def blend_f(x, ys):
            u = (x - x_blend_start) / x_blend_period
            return (1.0 - u) * ys[0] + u * ys[1]
        c = Aggregate([self, func], tfm=blend_f, name='blend')

        funcs = [self, c, func]
        domains = self.domain.partition([x_blend_start, x_blend_stop])
        return Piecewise(funcs, domains)

    def extension(self, name, start=False, end=True, raise_on_empty=False, **kwds):
        from .extension import ConstantExtension
        from .extension import TangentExtension
        from .extension import SinExtension
        from .extension import HarmonicExtension
        from .extension import WeaveExtension

        classes = [            
            ConstantExtension,
            TangentExtension,
            SinExtension,
            HarmonicExtension,
            WeaveExtension
        ]

        for c in classes:
            if c.name == name:
                return c(self, start=start, end=end, raise_on_empty=raise_on_empty, **kwds)
        
        raise Exception('Unknown extension type')

    # def wave_extended(self, ref_func, min_deviation=0, start=None, step=None, min_step=MIN_STEP):
    #     if self.domain.is_positive_infinite:
    #         return self
    #     ref_func = Curve.parse(ref_func)
    #     extremas = Extremas(self, ref_func, min_deviation=min_deviation, start=start, step=step, min_step=min_step)

    # def mom(self, degree, duration, **kwargs):
    #     """
    #     Returns the momentum of the reciever.

    #     The degree corresponds to the number of steps to take.
    #     """
    #     degree = int(degree)
    #     if degree < 1:
    #         raise ValueError(f'Momentum requires a positive degree, got: {degree}')
    #     from pyduration import Duration
    #     duration = Duration.parse(duration)
    #     def _mom(x, y):
    #         if y is None:
    #             return None
    #         # step back
    #         x0 = duration.step(x, -degree)
    #         y0 = self.y(x0)
    #         if y0 is None:
    #             return None
    #         return y - y0
    #     return self.map(_mom, name=f'mom({degree})', **kwargs)

    def sma(self, degree, is_period=False, **kwargs):
        from .sma import SMA
        return SMA(self, degree, is_period=is_period, **kwargs)

    def ema(self, degree, is_period=False, init=None, **kwargs):
        from .ema import EMA
        return EMA(self, degree, is_period=is_period, init=init, **kwargs)

    def smma(self, degree, **kwargs):
        from .sma import SMA
        from .ema import EMA
        sma = SMA(self, degree, is_period=False, **kwargs)
        ema = EMA(self, 1 / degree, is_period=False, init=sma, **kwargs)
        return ema

    def harmonic_smas(self, base_degree, count, stride=1, is_period=False, **kwargs):
        """
        Returns `count` SMAs from small to large. Their degrees
        are proportional to the golden ratio.
        """
        periods = []
        smas = []
        step = stride + 1

        for i in range(count):
            period = base_degree * GOLD ** float(i * step)
            period = round(period / base_degree) * base_degree
            periods.append(period)

        for i in range(count):
            period = periods[i]
            sma = self.sma(period, is_period=is_period, **kwargs)
            smas.append(sma)

        return smas

    def centered_macs(self, base_degree, count, stride=1, is_period=False, **kwargs):
        periods = []
        smas = []
        step = stride + 1

        for i in range(count):
            period = base_degree * GOLD ** float(i * step)
            period = round(period / base_degree) * base_degree
            periods.insert(0, period)

        for i in range(count):
            period = periods[i]
            sma = self.sma(period, is_period=is_period, **kwargs)
            smas.append(sma)

        return smas

    def rsi(self, degree, **kwargs):
        d = self.differential()
        du = Curve.max([d, 0], ignore_empty=False)
        dd = Curve.max([-d, 0], ignore_empty=False)
        rs = du.ema(1 / degree, **kwargs) / dd.ema(1 / degree, **kwargs)
        rsi = 100 - 100 / (1 + rs)
        rsi.name = f'rsi({degree})'
        return rsi

    def trailing_min(self, degree, is_period=False, interpolation=None, min_step=MIN_STEP, uniform=True):
        return self.accumulator_map(
            min,
            degree,
            is_period=is_period,
            interpolation=interpolation,
            min_step=min_step,
            uniform=uniform
        )

    def trailing_max(self, degree, is_period=False, interpolation=None, min_step=MIN_STEP, uniform=True):
        return self.accumulator_map(
            max,
            degree,
            is_period=is_period,
            interpolation=interpolation,
            min_step=min_step,
            uniform=uniform
        )

    def differential(self, forward=False):
        from .map import Map
        d = Map(self, lambda x, y: self.d_y(x, forward=forward))
        d.name = 'diff'
        return d

    def subset(self, domain):
        from .generic import Generic
        return Generic(self, domain=domain, min_step=self.min_step)

    @classmethod
    def first(cls, funcs, *args):
        """
        Return a func which returns the first value which is not `None`.
        """
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def first_val(x, vals):
            for v in vals:
                if v is not None:
                    return v
            return None

        funcs = Curve.parse_many(funcs)
        return Aggregate(funcs, tfm=first_val, union=True, name='first')        

    @classmethod
    def min(cls, funcs, *args, ignore_empty=False):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def min_vals(x, vals):
            best = None
            for val in vals:
                if best is None or (val is not None and val < best):
                    best = val
            return best

        def min_vals_with_empty(x, vals):
            return min(filter(lambda y: y is not None, vals), default=None)

        funcs = Curve.parse_many(funcs)
        t = min_vals_with_empty if ignore_empty else min_vals
        return Aggregate(funcs, tfm=t, union=ignore_empty, name='min')

    @classmethod
    def max(cls, funcs, *args, ignore_empty=False):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def max_vals(x, vals):
            best = None
            for val in vals:
                if best is None or (val is not None and val > best):
                    best = val
            return best

        def max_vals_with_empty(x, vals):
            return max(filter(lambda y: y is not None, vals), default=None)

        funcs = Curve.parse_many(funcs)
        t = max_vals_with_empty if ignore_empty else max_vals
        return Aggregate(funcs, tfm=t, union=ignore_empty, name='max')

    @classmethod
    def add_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def add_f(x, ys):
            for y in ys:
                if y is None:
                    return None
            return sum(ys)
        return Aggregate(funcs, tfm=add_f, name='add', operator='+')

    @classmethod
    def subtract_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def sub_f(x, ys):
            result = 0
            for i, y in enumerate(ys):
                if y is None:
                    return None
                if i == 0:
                    result = y
                else:
                    result -= y
            return result
        return Aggregate(funcs, tfm=sub_f, name='sub', operator='-')

    @classmethod
    def multiply_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def mult_f(x, ys):
            geo_sum = 1.0
            for y in ys:
                if y is None:
                    return None
                geo_sum *= y
            return geo_sum
        return Aggregate(funcs, tfm=mult_f, name='mult', operator='*')

    @classmethod
    def divide_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def div_f(x, ys):
            result = 0
            for i, y in enumerate(ys):
                if y is None:
                    return None
                if i == 0:
                    result = y
                elif y == 0:
                    result = math.inf if result >= 0 else -math.inf
                else:
                    result /= y
            return result
        return Aggregate(funcs, tfm=div_f, name='div', operator='/')

    @classmethod
    def pow_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def log_f(x, ys):
            result = 0
            for i, y in enumerate(ys):
                if y is None:
                    return None
                if i == 0:
                    result = y
                else:
                    result = result ** y
            return result
        return Aggregate(funcs, tfm=log_f, name='pow', operator='^')

    @classmethod
    def log_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        from .aggregate import Aggregate

        def log_f(x, ys):
            result = 0
            for i, y in enumerate(ys):
                if y is None:
                    return None
                if i == 0:
                    result = y
                else:
                    result = math.log(result, y)
            return result
        return Aggregate(funcs, tfm=log_f, name='log')

    @classmethod
    def zero(cls, value):
        from .constant import Constant
        return Constant.zero()

    @classmethod
    def const(cls, value):
        from .constant import Constant
        return Constant(value)

    @classmethod
    def parse(cls, func):
        from .generic import Generic
        from .constant import Constant
        from .points import Points

        if func is None:
            return None
        elif isinstance(func, Curve):
            return func
        elif callable(func):
            return Generic(func)
        elif isinstance(func, Number):
            return Constant(func)
        elif isinstance(func, Sequence):
            # Parse points
            if len(func) == 0:
                return Points(func)
            else:
                if isinstance(func[0], Sequence):
                    if len(func[0]) == 2:
                        return Points(func)
        elif isinstance(func, Mapping):
            return cls.parse_descriptor(func)
        raise Exception('Unable to parse function')

    @classmethod
    def parse_descriptor(cls, d, fragment=False, current_func=None, decorators=None):
        # Example:
        # {
        #     "$line": {
        #         "points": [
        #             ["2020-02-12 01:23+1200", 8765.56],
        #             ["2020-02-30 04:50+1200", 6765.56]
        #         ]
        #     }
        # }
        if decorators is None:
            decorators = []

        def next_func_constructor(fname):
            f = current_func or _func_obj
            assert isinstance(f, Curve)
            ftype = type(f)
            fconstructor = None
            fconstructor_from_instance = False
            type_method_names = list(map(lambda x: x[0], inspect.getmembers(ftype, predicate=inspect.ismethod)))
            f_method_names = list(map(lambda x: x[0], inspect.getmembers(f, predicate=inspect.ismethod)))
            if f'{fname}_many' in type_method_names:
                fname = f'{fname}_many'
            if fname in type_method_names:
                def _create_class_fconstructor(fname):
                    def _class_fconstructor(*args, **kwargs):
                        f = current_func or _func_obj
                        fmethod = getattr(type(f), fname)
                        return fmethod(*args, **kwargs)
                    return _class_fconstructor
                fconstructor = _create_class_fconstructor(fname)
                fconstructor_from_instance = False
            elif fname in f_method_names:
                def _create_fconstructor(fname):
                    def _fconstructor(*args, **kwargs):
                        f = current_func or _func_obj
                        fmethod = getattr(f, fname)
                        return fmethod(*args, **kwargs)
                    return _fconstructor
                fconstructor = _create_fconstructor(fname)
                fconstructor_from_instance = True
            else:
                raise ValueError(f'Bad function name: {fname}')
            return fconstructor, fconstructor_from_instance
        
        if isinstance(d, Mapping):
            fragment_vals = {}

            for k, v in d.items():

                if k.startswith('@'):
                    # This is an decorator descriptor
                    oname = k[1:]
                    decorator_i = len(decorators)
                    decorators.insert(decorator_i, oname)
                    v = cls.parse_descriptor(v,
                        fragment=fragment,
                        current_func=current_func,
                        decorators=decorators
                    )
                    del decorators[decorator_i]

                    if oname.startswith('log'):
                        # Log space has ended, exit log space
                        # by raising to power
                        base_str = oname[3:]
                        base = int(base_str) if bool(base_str) else math.e
                        v = base ** v
                    
                    if isinstance(v, Curve):
                        # Allow chaining
                        current_func = v
                        continue
                    if oname != 'args':
                        # Only let @args pass through to parent
                        if len(d) != 1:
                            raise ValueError(f'A decorator (@...) can only have siblings in a fragment')
                        return v

                elif k.startswith('$'):
                    # This is function descriptor
                    fname = k[1:]

                    fconstructor = None
                    fconstructor_from_instance = False

                    if fname == 'const' or fname == 'constant':
                        from .constant import Constant
                        fconstructor = Constant
                    elif fname == 'line':
                        from .line import Line
                        fconstructor = Line
                    elif fname.startswith('log'):
                        base_str = fname[3:]
                        base = int(base_str) if bool(base_str) else math.e
                        def _dot_log(*args, **kwargs):
                            return current_func.log(**util.extend({ "base": base }, kwargs))
                        fconstructor = _dot_log
                        fconstructor_from_instance = True
                    else:
                        fconstructor, fconstructor_from_instance = next_func_constructor(fname)

                    func_args = cls.parse_descriptor(v,
                        fragment=True,
                        decorators=decorators
                    )
                    args = []
                    kwargs = {}
                    if isinstance(func_args, dict):
                        kwargs = func_args
                    elif isinstance(func_args, list):
                        args = func_args
                    else:
                        args = [func_args]
                    
                    # Check for nested args
                    if '@args' in kwargs:
                        args = kwargs['@args']
                        del kwargs['@args']

                    if fconstructor_from_instance and current_func is None:
                        current_func = Curve.parse(args[0])
                        del args[0]
                    elif not fconstructor_from_instance and current_func is not None:
                        # Add current function as first argument or to
                        # list at first argument
                        if bool(args) and isinstance(args[0], list):
                            args[0][0:0] = [current_func]
                        else:
                            args[0:0] = [current_func]
                    current_func = fconstructor(*args, **kwargs)
                    continue

                if current_func is not None:
                    raise Exception(f'Unexpected key after a function: {k}')

                if isinstance(v, Mapping):
                    fragment_vals[k] = cls.parse_descriptor(v,
                        fragment=True,
                        decorators=decorators
                    )
                elif isinstance(v, Sequence) and not isinstance(v, (str, bytes)):
                    fragment_vals[k] = cls.parse_descriptor(v,
                        fragment=True,
                        decorators=decorators
                    )
                else:
                    fragment_vals[k] = v
            
            return current_func or fragment_vals
        elif fragment:
            if isinstance(d, Mapping):
                return {k: cls.parse_descriptor(v,
                    fragment=True,
                    decorators=decorators
                ) for k, v in d.items()}
            elif isinstance(d, Sequence) and not isinstance(d, (str, bytes)):
                return [cls.parse_descriptor(v,
                    fragment=True,
                    decorators=decorators
                ) for v in d]
            elif 'date' in decorators:
                return arrow.get(d).timestamp
            elif isinstance(d, Number):
                if 'log' in decorators:
                    return math.log(d)
                elif 'log2' in decorators:
                    return math.log(d, 2)
                elif 'log10' in decorators:
                    return math.log(d, 10)
                else:
                    return d
            else:
                return d
        else:
            raise TypeError('Unexpected type found while parsing a function')

    @classmethod
    def parse_many(cls, funcs, *args):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        return list(map(cls.parse, funcs))

    @classmethod
    def count_positional_args(cls, f, default=1):
        if not callable(f):
            raise Exception('Expected callable function')
        if inspect.isbuiltin(f):
            return default
        sig = inspect.signature(f)
        count = 0
        for param in sig.parameters.values():
            if param.kind == inspect.Parameter.POSITIONAL_ONLY or \
                param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                count += 1
        return count

    def __add__(self, other):
        return Curve.add_many([self, other])

    def __sub__(self, other):
        return Curve.subtract_many([self, other])

    def __mul__(self, other):
        return Curve.multiply_many([self, other])

    def __truediv__(self, other):
        return Curve.divide_many([self, other])

    def __pow__(self, other):
        return Curve.pow_many([self, other])

    def __radd__(self, other):
        return Curve.add_many([other, self])

    def __rsub__(self, other):
        return Curve.subtract_many([other, self])

    def __rmul__(self, other):
        return Curve.multiply_many([other, self])

    def __rtruediv__(self, other):
        return Curve.divide_many([other, self])

    def __rpow__(self, other):
        return Curve.pow_many([other, self])

    def __neg__(self):
        return self.additive_inverse()

    def __pos__(self):
        return self

    def __abs__(self):
        return self.abs()

def _additive_inverse(x, y):
    if y is None:
        return None
    return -y

def _multiplicative_inverse(x, y):
    if y is None:
        return None
    return 1 / y

def _abs(x, y):
    if y is None:
        return None
    return abs(y)

def _callable_arg_len(f, vararg_ret_val):
    args, varargs, _, _ = inspect.getargspec(f)
    if varargs is not None:
        return vararg_ret_val
    arg_len = len(args)
    if arg_len == 0:
        return 0
    if args[0] == 'self':
        arg_len -= 1
    return arg_len

_func_obj = Curve()



import math
from .accumulator import Accumulator
from .curve import Curve, MIN_STEP
from intervalpy import Interval

class EMA(Accumulator):
    """
    Only forward direction is supported.
    """

    # TODO: investigate difference between regular EMA and irregular EMA

    def __init__(self, func, degree, is_period=True, init=None, min_step=MIN_STEP, uniform=True):
        self.alpha = None
        self.period = None
        self.init_func = Curve.parse(init) if init is not None else None
        if is_period:
            self.period = degree
        else:
            self.alpha = degree
        super().__init__(func, self._ema_scan, min_step=min_step, uniform=uniform)

    def __repr__(self):
        try:
            return f'{self.curve}.ema({self.alpha or self.period})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    # TODO: EMA doesn't have to initialise from the start of the underlying function
    # def init_scan(self, x):
    #     if self.period is not None:
    #         period = self.period
    #     elif self.alpha is not None:
    #         period = 1 / self.alpha
    #     else:
    #         raise Exception('Bad SMA configuration')       
    #     return self.curve.x_next(x - period, min_step=self.min_step)

    def _ema_scan(self, x, y, ema):
        if y is None:
            return ema
        if ema is None:
            if self.init_func is not None:
                return self.init_func.y(x)
            else:
                return y
        if self.period is not None:
            # TODO: use geometric mean instead of arithmetic?
            x0 = self.curve.x_previous(x, min_step=self.min_step)
            a = abs((x - x0) / self.period)
            return ema + a * (y - ema)
        elif self.alpha is not None:
            return ema + self.alpha * (y - ema)


# # Reference: http://www.thalesians.com/archive/public/academic/finance/papers/Zumbach_2000.pdf
# x0 = self.x_previous(x, min_step=self.min_step)
# y0 = self.curve(x0)
# x_delta = abs(x - x0)
# a = abs(x_delta / self.period)
# u = math.exp(-a)
# v = (1 - u) / a
# return u * ema + (v - u) * y0 + (1 - v) * y


import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval
from pyduration import Duration


class Offset(Curve):

    """
    Offset a function on the x-axis by a constant.
    """

    def get_domain(self):
        return self._offset_interval(self.curve.domain)

    def __init__(self, func, offset, duration=None):
        super().__init__()
        self.curve = Curve.parse(func)
        self.offset = offset
        self.duration = None
        if duration is not None:
            self.duration = Duration.parse(duration)
            if type(self.offset) != int:
                raise Exception('Offset must be an interger when duration is defined')
        self._observer_token = self.curve.add_observer(
            begin=self.begin_offset_update, end=self.end_offset_update, prioritize=True)

    def __del__(self):
        self.curve.remove_observer(self._observer_token)

    def __repr__(self):
        try:
            return f'{self.curve}.offset({self.offset})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def y(self, x):
        return self._interpolated_func(self.curve.y, x)

    def d_y(self, x, **kwargs):
        return self._interpolated_func(self.curve.d_y, x, **kwargs)

    def _interpolated_func(self, f, x, **kwargs):
        x0 = self._unoffset_x(x, floor=True)
        x1 = self._unoffset_x(x, floor=False)
        if x0 is None or x1 is None or not self.curve.domain.contains(x0) or not self.curve.domain.contains(x1):
            return None
        if x0 == x1:
            return f(x0, **kwargs)
        y0 = f(x0, **kwargs)
        y1 = f(x1, **kwargs)

        if y0 is None or y1 is None:
            return None
        
        _x0 = self._offset_x(x0, floor=True)
        _x1 = self._offset_x(x1, floor=False)
        u = (x - _x0) / (_x1 - _x0)
        y = y0 * (1 - u) + y1 * u
        return y

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        x = self._unoffset_x(x, floor=False)
        if x is None:
            return None
        if self.duration:
            x1 = self.duration.previous(x)
        else:
            x1 = self.curve.x_previous(x, min_step=min_step, limit=limit)
        if x1 is None or not self.curve.domain.contains(x1):
            return None
        return self._offset_x(x1, floor=False)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        x = self._unoffset_x(x, floor=True)
        if x is None:
            return None
        if self.duration:
            x1 = self.duration.next(x)
        else:
            x1 = self.curve.x_next(x, min_step=min_step, limit=limit)
        if x1 is None or not self.curve.domain.contains(x1):
            return None
        return self._offset_x(x1, floor=True)

    def begin_offset_update(self, domain):
        self.begin_update(self._offset_interval(domain))

    def end_offset_update(self, domain):
        self.end_update(self._offset_interval(domain))

    def _offset_x(self, x, floor=False):
        if x is None or math.isinf(x):
            return x
        if self.duration is None:
            return x + self.offset
        else:
            return self._res_step(x, self.offset, floor=floor)

    def _unoffset_x(self, x, floor=False):
        if x is None or math.isinf(x):
            return x
        if self.duration is None:
            return x - self.offset
        else:
            return self._res_step(x, -self.offset, floor=floor)

    def _res_step(self, x, count, floor=False):
        if count == 0:
            return x
        if floor:
            x = self.duration.floor(x)
        else:
            x = self.duration.ceil(x)
        return self.duration.step(x, count=count)

    def _offset_interval(self, domain, floor=False):
        start = self._offset_x(domain.start, floor=floor)
        end = self._offset_x(domain.end, floor=floor)
        if start is None and end is None:
            return Interval.empty()
        elif start is None:
            return Interval(start, end, start_open=False, end_open=domain.end_open)
        elif end is None:
            return Interval(start, end, start_open=domain.start_open, end_open=False)
        else:
            return Interval(start, end, start_open=domain.start_open, end_open=domain.end_open)



from .curve import Curve, MIN_STEP
from intervalpy import Interval
from collections.abc import Sequence


class Aggregate(Curve):

    def get_domain(self):
        domains = list(map(lambda f: f.domain, self.funcs))
        if self.is_union:
            return Interval.union(domains)
        else:
            return Interval.intersection(domains)

    def __init__(self, funcs, *args, tfm=None, default=None, union=False, operator=None, name=None):
        if not isinstance(funcs, Sequence):
            funcs = [funcs] + list(args)
        super().__init__()
        self.funcs = Curve.parse_many(funcs)
        self.tfm = tfm
        self.is_union = union
        self.operator = operator
        self.name = name
        self.default = default
        unique_funcs = list(set(list(self.funcs)))
        self._observer_tokens = list(map(lambda f: f.add_observer(
            begin=self.begin_update, end=self.end_update, prioritize=True), unique_funcs))

    def __del__(self):
        for func, token in zip(self.funcs, self._observer_tokens):
            func.remove_observer(token)

    def __repr__(self):
        try:
            params = ", ".join(list(map(str, self.funcs)))
            return f'{self.name or self.operator or self.util.__name__}({params})'
            # if self.operator is not None:
            #     def child_str(f):
            #         if isinstance(f, Aggregate):
            #             return f'({f})'
            #         else:
            #             return str(f)
            #     return f" {self.operator} ".join(list(map(child_str, self.funcs)))
            # else:
            #     params = ", ".join(list(map(str, self.funcs)))
            #     return f'{self.name or self.util.__name__}({params})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def y(self, x):
        if not self.domain.contains(x):
            return self.default
        func_vals = list(map(lambda f: f.y(x), self.funcs))
        if self.tfm is None:
            return func_vals
        return self.tfm(x, func_vals)

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        return max_or_none(map(lambda f: f.x_previous(x, min_step=min_step, limit=limit), self.funcs), x)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        return min_or_none(map(lambda f: f.x_next(x, min_step=min_step, limit=limit), self.funcs), x)


def min_or_none(iterable, gt):
    min = None
    for item in iterable:
        if item is None:
            continue
        if min is None or item < min:
            if item > gt:
                min = item
    return min


def max_or_none(iterable, lt):
    max = None
    for item in iterable:
        if item is None:
            continue
        if max is None or item > max:
            if item < lt:
                max = item
    return max



from .const import GOLD
from .curve import Curve, MIN_STEP
from .extension.tangent import TangentExtension
from .extension.constant import ConstantExtension
from .accumulator import Accumulator
from .aggregate import Aggregate
from .constant import Constant
from .ema import EMA
from .empty import Empty
from .extremas import Extremas
from .generic import Generic
from .integral import Integral
from .line import Line
from .map import Map
from .offset import Offset
from .piecewise import Piecewise
from .points import Points
from .quotes import Quotes
from .scan import Scan
from .sin import Sin
from .sma import SMA
from .trend import Trend
from .accumulator_map import AccumulatorMap



import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval


class Line(Curve):

    def get_domain(self):
        return Interval.infinite()

    def __init__(self, const=None, slope=None, p1=None, p2=None, points=None):
        super().__init__(min_step=math.inf)
        self.ref_point = (0, 0)
        self.slope = 0
        self.set(const=const, slope=slope, p1=p1, p2=p2, points=points)

    @property
    def const(self):
        return self.y(0)

    def __repr__(self):
        try:
            if self.ref_point[0] == 0:
                x = 'x'
            else:
                x = f'(x - {self.ref_point[0]})'
            if self.ref_point[1] == 0:
                _y = ''
            else:
                _y = f' + {self.ref_point[1]}'

            return f'{x} * {self.slope}{_y}'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def set(self, const=None, slope=None, p1=None, p2=None, points=None):
        ref_point = (0, const)
        slope = slope
        if points is not None:
            p1 = points[0]
            p2 = points[1]
        if p1 is None:
            p1 = ref_point
        else:
            ref_point = (p1[0], p1[1])
        if p2 is not None:
            if p2[0] != p1[0]:
                slope = (p2[1] - p1[1]) / (p2[0] - p1[0])
            elif p2[1] >= p1[1]:
                slope = math.inf
            else:
                slope = -math.inf
        if ref_point[0] is None or ref_point[1] is None:
            raise Exception("Line requires a constant or reference point")
        if slope is None:
            raise Exception("Line requires a slope")

        if ref_point[0] == self.ref_point[0] and ref_point[1] == self.ref_point[1] and slope == self.slope:
            return

        self.begin_update(self.domain)
        self.ref_point = ref_point
        self.slope = slope
        self.end_update(self.domain)

    def y(self, x):
        return self.ref_point[1] + self.slope * (x - self.ref_point[0])

    def d_y(self, x, forward=False, min_step=MIN_STEP, limit=None):
        return self.slope

    def x(self, y):
        return self.ref_point[0] + (y - self.ref_point[1]) / self.slope

    def partial_integration(self, a, b):
        # f(x) = slope / 2 * x ^ 2 + const * x + K
        # integral = f(b) - f(a)
        if a is not None and a == b:
            return 0
        const = self.const
        slope = self.slope
        ia = slope * 0.5 * a ** 2 + const * a
        ib = slope * 0.5 * b ** 2 + const * b
        return ib - ia



from ..curve import Curve, MIN_STEP
from intervalpy import Interval

class Extension(Curve):

    """
    Extends ends of a function.
    """

    name = ""

    def get_domain(self):
        self._update_extension_interval()
        return Interval.union([self.curve.domain, self.extension_interval])

    def __init__(self, func, start=True, end=True, uniform=True, raise_on_empty=False, min_step=MIN_STEP):
        super().__init__(min_step=min_step)
        self.curve = Curve.parse(func)

        if self.curve.domain.is_negative_infinite:
            start = False
        if self.curve.domain.is_positive_infinite:
            end = False

        self.start = start
        self.end = end
        self.start_valid = True
        self.end_valid = True
        self.uniform = uniform
        self.raise_on_empty = raise_on_empty
        self.extension_interval = Interval.empty()
        self._extension_stale = True
        self._extension_interval_stale = True
        self.curve.add_observer(begin=self.begin_extension_update, end=self.end_extension_update, prioritize=True)
        self.start_func = None
        self.end_func = None
        if self.start:
            self.start_func = self.create_extension_func(start=True)
            self.start_func.add_observer(self, begin=self.begin_update, end=self.end_update)
        if self.end:
            self.end_func = self.create_extension_func(start=False)
            self.end_func.add_observer(self, begin=self.begin_update, end=self.end_update)

    def __del__(self):
        self.curve.remove_observer(self)

    def __repr__(self):
        try:
            vals = [type(self).__name__]
            if self.start:
                vals.append('start')
            if self.end:
                vals.append('end')
            return f'{self.curve}.extension({", ".join(vals)})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def update_extension(self):
        """
        Subclasses should override this method and refresh the extension.
        """
        pass

    def update_extension_interval(self):
        """
        Subclasses should override this method and refresh the extension domain if needed.
        """
        pass

    def create_extension_func(self, start=False):
        raise Exception('Not implemented')

    def y_extension(self, x):
        """
        Return the value of the extension for a given `x` value.
        """
        if self.curve.domain.is_empty:
            return None
        if self.start and x <= self.curve.domain.start:
            return self.start_func.y(x)
        if self.end and x >= self.curve.domain.end:
            return self.end_func.y(x)
        return None

    def y(self, x):
        if self.curve.domain.contains(x):
            return self.curve.y(x)
        self._update_extension_if_needed()
        if self.extension_interval.contains(x):
            return self.y_extension(x)
        return None

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        x1 = self.curve.x_previous(x, min_step=min_step)
        if x1 is None and self.start:
            # Keep returning extension with same step
            x0 = self.curve.x_next(self.curve.domain.start, min_step=min_step, limit=limit)
            if x0 is None:
                return None
            step = x0 - self.curve.domain.start
            x1 = x - step
        return x1

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        x1 = self.curve.x_next(x, min_step=min_step)
        if x1 is None and self.end:
            # Keep returning extension with same step
            x0 = self.curve.x_previous(self.curve.domain.end, min_step=min_step, limit=limit)
            if x0 is None:
                return None
            step = self.curve.domain.end - x0
            x1 = x + step
        return x1

    def is_in_extension(self, x):
        domain = Interval.parse(x)
        return (self.start and domain <= self.extension_interval) or \
            (self.end and domain >= self.extension_interval)

    def begin_extension_update(self, domain):
        if self.is_in_extension(domain):
            self.begin_update(self.extension_interval)
        else:
            self.begin_update(domain)

    def end_extension_update(self, domain):
        if self.is_in_extension(domain):
            self._extension_stale = True
            self._update_extension_interval()
            self.end_update(self.extension_interval)
        else:
            self.end_update(domain)

    def _update_extension_interval(self):
        if not self._extension_interval_stale:
            return
        self._extension_interval_stale = False
        if self._extension_stale:
            self._update_extension_if_needed()

        if self.start and self.start_valid and self.end and self.end_valid:
            self.extension_interval = Interval.union([self.start_func.domain, self.end_func.domain])
        elif self.start and self.start_valid:
            self.extension_interval = Interval.intersection([self.start_func.domain, self.curve.domain.get_lt()])
        elif self.end and self.end_valid:
            self.extension_interval = Interval.intersection([self.end_func.domain, self.curve.domain.get_gt()])
        else:
            self.extension_interval = Interval.empty()
        self.update_extension_interval()
        if (self.start and not self.extension_interval.is_negative_infinite) or (self.end and not self.extension_interval.is_positive_infinite):
            if self.raise_on_empty:
                raise Exception('Unable to extend func')
            
    def _update_extension_if_needed(self):
        if not self._extension_stale:
            return
        self._extension_stale = False
        self._extension_interval_stale = True

        self.update_extension()
        if (self.start and not self.extension_interval.is_negative_infinite) or (self.end and not self.extension_interval.is_positive_infinite):
            if self.raise_on_empty:
                raise Exception('Unable to extend func')
        self._update_extension_interval()



import math
from .curve import Curve, MIN_STEP
from .points import Points
from .integral import Integral
from intervalpy import Interval

class SMA(Integral):

    """
    Only forward direction is supported.
    """

    def get_domain(self):
        d = self.curve.domain
        if d.is_empty:
            return Interval.empty()
        if self.period is not None:
            x = self.curve.x_previous(d.start + self.period, min_step=self.min_step)
            if x is None:
                p = self.period
            else:
                p = x - d.start
        elif self.degree is not None:
            x = d.start
            for _ in range(self.degree - 1):
                x = self.curve.x_next(x, min_step=self.min_step)
                if x is None:
                    return Interval.empty()
            p = x - d.start
        else:
            raise Exception('Bad SMA configuration')
        d_start = d.start + p
        if math.isnan(d_start) or d_start > d.end:
            return Interval.empty()
        return Interval.closed(d_start, d.end)

    def __init__(self, func, degree, is_period=True, uniform=True, min_step=MIN_STEP):
        super().__init__(func, uniform=uniform, min_step=min_step)
        self.degree = None
        self.period = None
        if is_period:
            self.period = degree
            if self.period is None or self.period <= 0:
                raise Exception('SMA period must be a positive number')
        else:
            self.degree = int(degree)
            if self.degree <= 0:
                raise Exception('SMA degree must be a positive integer')

    def __repr__(self):
        try:
            return f'{self.curve}.sma({self.degree or self.period})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def init_scan(self, x):
        return x
        # if self.curve.domain.is_empty or self.curve.domain.is_negative_infinite:
        #     return x
        # x0 = self.curve.x_previous(x, min_step=self.min_step)
        # if not self.domain.contains(x0):
        #     x0 = self.curve.x_next(self.curve.domain.start, min_step=self.min_step)
        # return x0

    def offset_scan(self, x):
        return self._sma_start(x)

    def y(self, x):
        # Return the average slope as the SMA
        if not self.domain.contains(x):
            return None

        # Get step
        x0 = self._sma_start(x)
        x1 = self._sma_end(x0)
        if x0 == x or x0 == x1:
            # SMA step is smaller than or equal to the underlying step
            return self.curve.y(x)

        # Values must be accessed in ascending order
        y0 = super().y(x0)
        if y0 is None:
            return None
        y1 = super().y(x1)
        if y1 is None:
            return None

        sma1 = (y1 - y0) / (x1 - x0)
        if x == x1:
            return sma1

        # Handle offset
        if x > x1:
            x2 = self.curve.x_next(x1, min_step=self.min_step)
            if x2 == x:
                return sma1
            sma2 = self.y(x2)
            if sma2 is None:
                return None
            u = (x - x1) / (x2 - x1)
            sma = u * sma2 + (1 - u) * sma1
        else:
            x01 = self.curve.x_previous(x1, min_step=self.min_step)
            if x01 == x:
                return sma1
            sma01 = self.y(x01)
            if sma01 is None:
                return None
            u = (x - x01) / (x1 - x01)
            sma = u * sma01 + (1 - u) * sma1
            return sma
        return sma

    def _sma_start(self, x):
        if self.period is not None:
            x0 = self.curve.x_next(x - self.period, min_step=self.min_step)
        elif self.degree is not None:
            x0 = x
            for _ in range(0, self.degree - 1):
                x1 = self.curve.x_previous(x0, min_step=self.min_step)
                if x1 is None:
                    return x0
                x0 = x1
        else:
            raise Exception('Bad SMA configuration')
        return x0

    def _sma_end(self, x):
        if self.period is not None:
            x0 = self.curve.x_previous(x + self.period, min_step=self.min_step)
        elif self.degree is not None:
            x0 = x
            for _ in range(0, self.degree - 1):
                x1 = self.curve.x_next(x0, min_step=self.min_step)
                if x1 is None:
                    return x0
                x0 = x1
        else:
            raise Exception('Bad SMA configuration')
        return x0



import bisect
from .scan import Scan
from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Extremas(Scan):

    # TODO: Extremas doesn't need to be a subclass of `Scan` as it only needs to find the local extrema about a point.

    def __init__(self, func, ref_func, min_deviation=0, min_step=MIN_STEP):
        self.ref_func = Curve.parse(ref_func)
        self.min_deviation = abs(min_deviation)
        self.extremas = []
        self.extrema_xs = []
        self.extrema_interval = Interval.empty()
        self.possible_extrema = None
        self.possible_extrema_phase = None
        self._did_update_extremas()
        super().__init__(func, self._extrema_scan, min_step=min_step)
        self._ref_observer_token = self.ref_func.add_observer(begin=self.begin_update, end=self.end_update)

    def __del__(self):
        self.ref_func.remove_observer(self._ref_observer_token)

    def scanned_y(self, x):
        if not self.extrema_interval.contains(x):
            return None
        if x == self.extrema_interval.start:
            return self.extremas[0][1]
        elif x == self.extrema_interval.end:
            return self.extremas[-1][1]
        i = bisect.bisect(self.extrema_xs, x)
        p1 = self.extremas[i - 1]
        p2 = self.extremas[i]
        u = (x - p1[0]) / (p2[0] - p1[0])
        return (1 - u) * p1[1] + u * p2[1]

    def continue_scan(self, x):
        if super().continue_scan(x):
            return True
        if not self.domain.contains(self.current):
            return False
        return not self.extrema_interval.contains(x, enforce_start=False)

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        if self.extrema_interval.contains(x - min_step):
            i = bisect.bisect_left(self.extrema_xs, x - min_step)
            x1 = self.extremas[i][0]
            if x1 > x - min_step:
                x1 = self.extremas[i - 1][0]
            return x1
        return self.curve.x_previous(x, min_step=min_step, limit=limit)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        if self.extrema_interval.contains(x + min_step):
            i = bisect.bisect_left(self.extrema_xs, x + min_step)
            x1 = self.extremas[i][0]
            if x1 < x + min_step:
                x1 = self.extremas[i + 1][0]
            return x1
        return self.curve.x_next(x, min_step=min_step, limit=limit)

    def begin_update(self, domain):
        super().begin_update(domain)
        # remove stale points
        for i in reversed(range(len(self.extremas))):
            x = self.extrema_xs[i]
            if domain.contains(x):
                self._remove_extrema_index(i)
            elif domain.start > x:
                break

        # scan from last extrema
        extrema_count = len(self.extremas)
        if extrema_count == 0:
            self.current = None
        else:
            last_extrema = self.extremas[-1]
            x = last_extrema[0]
            self.current = x
            self.possible_extrema = last_extrema
            self.possible_extrema_phase = last_extrema[1] - self.ref_func(x)

    def sample_points(self, domain=None, min_step=MIN_STEP, step=None):
        points = super().sample_points(domain=domain, min_step=min_step, step=step)
        return list(filter(lambda p: p[1] is not None, points))

    def _extrema_scan(self, x, y):
        if y is None:
            return
        y0 = self.ref_func(x)
        if y0 is None:
            return
        div = y / y0 - 1
        if abs(div) <= self.min_deviation:
            # function is too close to reference function
            return
        phase = y - y0
        if self.possible_extrema is None:
            self.possible_extrema = (x, y)
            self.possible_extrema_phase = phase
            return
        is_possible_extrema = False
        if self.possible_extrema_phase * phase < 0:
            # phase inflection
            self._confirm_extrema()
            is_possible_extrema = True
        elif (phase > 0 and y > self.possible_extrema[1]) or (phase < 0 and y < self.possible_extrema[1]):
            is_possible_extrema = True
        if is_possible_extrema:
            self.possible_extrema = (x, y)
            self.possible_extrema_phase = phase
    
    def _remove_extrema_index(self, i):
        del self.extremas[i]
        del self.extrema_xs[i]
        self._did_update_extremas()

    def _confirm_extrema(self):
        extrema = self.possible_extrema

        self.extremas.append(extrema)
        self.extrema_xs.append(extrema[0])
        self._did_update_extremas()

    def _did_update_extremas(self):
        self.possible_extrema = None
        self.possible_extrema_phase = None

        if len(self.extremas) == 0:
            self.extrema_interval = Interval.empty()
        else:
            self.extrema_interval = Interval(self.extrema_xs[0], self.extrema_xs[-1], start_open=False, end_open=False)

    def _i(self, x):
        if not self.extrema_interval.contains(x):
            return None
        if x == self.extrema_interval.start:
            return 0
        elif x == self.extrema_interval.end:
            return len(self.extremas) - 1
        i = bisect.bisect(self.extrema_xs, x)
        p1 = self.extremas[i - 1]
        p2 = self.extremas[i]
        u = (x - p1[0]) / (p2[0] - p1[0])
        return i + u



import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Empty(Curve):

    def get_domain(self):
        return Interval.empty()

    def __init__(self):
        super().__init__(min_step=math.inf)

    def __repr__(self):
        return 'empty'

    def y(self, x):
        return None

    def d_y(self, x, forward=False, min_step=MIN_STEP, limit=None):
        return None

    def x(self, y):
        return None
        


import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Sin(Curve):

    @property
    def frequency(self):
        return self.x_coef / math.pi * 0.5

    @property
    def period(self):
        return math.pi * 2 / self.x_coef

    def get_domain(self):
        return Interval.infinite()

    def __init__(self, amplitude=None, frequency=None, period=None, phase=0, phase_x=None):
        super().__init__()
        self.set(amplitude=amplitude, frequency=frequency, period=period, phase=phase, phase_x=phase_x)
    
    def set(self, amplitude=None, frequency=None, period=None, phase=0, phase_x=None):
        self.begin_update(self.domain)
        self.amplitude = amplitude
        self.phase = phase

        self.x_coef = None
        if period is not None:
            frequency = 1.0 / period
        if frequency is not None:
            self.x_coef = frequency * math.pi * 2.0
        if phase_x is not None:
            self.phase = phase_x * self.x_coef

        if self.amplitude is None or self.x_coef is None or self.phase is None:
            raise Exception("Invalid sin function description")
        self.end_update(self.domain)

    def y(self, x):
        return self.amplitude * math.sin(self.x_coef * x + self.phase)

    def d_y(self, x, forward=False, min_step=MIN_STEP, limit=None):
        return self.amplitude * math.cos(self.x_coef * x + self.phase)
        


import numpy as np
import math
import copy
# from scipy.optimize import minimize
from collections.abc import Sequence
from .scan import Scan
from .curve import MIN_STEP
from intervalpy import Interval
from .const import GOLD, GOLD_2

TOL_REL_DEFAULT = 0.01
SEARCH_WIDTH_REL_DEFAULT = 10.0
SEARCH_WIDTH_INT_REL_DEFAULT = 2.0
EPSILON_ZERO = 1e-14

MAX_COUNT = 1000

class Trend(Scan):
    """
    Only forward direction is supported.
    """

    def get_domain(self):
        return super().get_domain().extended_to_positive_infinity()

    def __init__(self, func, min_width, min_points=3, min_point_distance=0, search_length_rel=SEARCH_WIDTH_REL_DEFAULT, search_length_int_rel=SEARCH_WIDTH_INT_REL_DEFAULT, x_tol_rel=None, x_tol_abs=None, y_tol_rel=None, y_tol_abs=None, min_step=MIN_STEP):
        self.min_width = min_width
        self.min_points = min_points
        self.min_point_distance = min_point_distance
        self.search_length_rel = search_length_rel
        self.search_length_int_rel = search_length_int_rel

        if x_tol_rel is None and x_tol_abs is None:
            x_tol_rel = TOL_REL_DEFAULT
        if y_tol_rel is None and y_tol_abs is None:
            y_tol_rel = TOL_REL_DEFAULT

        self.x_tol_rel = x_tol_rel
        self.x_tol_abs = x_tol_abs
        self.y_tol_rel = y_tol_rel
        self.y_tol_abs = y_tol_abs

        self.points = []
        self.nested_upper_lines = [[]]
        self.nested_lower_lines = [[]]
        
        super().__init__(func, self._trend_scan, min_step=min_step)

    @property
    def lines(self):
        return self._scanned_lower(None) + self._scanned_upper(None)

    def upper(self, x):
        if x is not None:
            self.scan(x)
        return self._scanned_upper(x)

    def lower(self, x):
        if x is not None:
            self.scan(x)
        return self._scanned_lower(x)

    def scanned_y(self, x):
        return self._scanned_lower(x) + self._scanned_upper(x)

    def continue_scan(self, x):
        if super().continue_scan(x):
            return True
        if not self.domain.contains(self.current):
            return False

        is_empty = True

        # for line in self.lines:
        #     is_empty = False
        #     if line.search_interval.contains(x, enforce_end=False):
        #         return True

        for lines in self.nested_upper_lines:
            for line in lines:
                is_empty = False
                if line.search_interval.contains(x, enforce_end=False):
                    return True

        for lines in self.nested_lower_lines:
            for line in lines:
                is_empty = False
                if line.search_interval.contains(x, enforce_end=False):
                    return True

        return is_empty

    def begin_update(self, domain):
        super().begin_update(domain)

        # remove stale points
        for i in reversed(range(len(self.points))):
            p = self.points[i]
            if domain.contains(p[0]):
                del self.points[i]
            elif p[0] < domain.start:
                break

        def update_nested_line_array(nested_lines):
            for i in reversed(range(len(nested_lines))):
                lines = nested_lines[i]
                update_line_array(lines)
                if i != 0 and len(lines) == 0:
                    del nested_lines[i]

        def update_line_array(lines):
            for i in reversed(range(len(lines))):
                line = lines[i]
                line.begin_update(domain, self.points)
                if line.is_empty:
                    del lines[i]

        # update_line_array(self.lines)
        update_nested_line_array(self.nested_upper_lines)
        update_nested_line_array(self.nested_lower_lines)

    def _scanned_upper(self, x):
        filtered_lines = []
        for lines in self.nested_upper_lines:
            for line in lines:
                if self._line_filter(line, x):
                    filtered_lines.append(line)
        return filtered_lines

    def _scanned_lower(self, x):
        filtered_lines = []
        for lines in self.nested_lower_lines:
            for line in lines:
                if self._line_filter(line, x):
                    filtered_lines.append(line)
        return filtered_lines

    def _line_filter(self, line, x):
        if not line.is_valid:
            return False
        if x is not None and not line.search_interval.contains(x):
            return False
        if self.min_width is not None and line.width < self.min_width:
            return False
        return True

    def _trend_scan(self, x, y):
        if y is None:
            return
        p = (x, y)
        self.points.append(p)
        
    #     self._did_append_nested_point(p, self.nested_upper_lines, is_upper_trend=True)
    #     self._did_append_nested_point(p, self.nested_lower_lines, is_upper_trend=False)

    # def _did_append_nested_point(self, nested_p, nested_lines, is_upper_trend=True):
        i = 0
        points = [p]
        while i < len(self.nested_upper_lines):
            all_trend_points = []
            for p in points:
                new_trend_points = self._append_point_and_validate(p, self.nested_upper_lines[i], self.nested_upper_lines, is_upper_trend=True)
                _insort_points_left(all_trend_points, new_trend_points)

                new_trend_points = self._append_point_and_validate(p, self.nested_lower_lines[i], self.nested_lower_lines, is_upper_trend=False)
                _insort_points_left(all_trend_points, new_trend_points)
            if len(all_trend_points) == 0:
                break
            # print('found {} trend line(s) at {} (depth {}/{})'.format('upper' if is_upper_trend else 'lower', nested_p, i, len(nested_lines)))
            # print(list(map(lambda l: l.trend_points, nested_lines[i])))
            i += 1
            points = all_trend_points
            if len(self.nested_upper_lines) == i:
                self.nested_upper_lines.append([])
                self.nested_lower_lines.append([])

    def _append_point_and_validate(self, p, lines, nested_lines, is_upper_trend=True):
        remove_indexes = []
        new_trend_points = []

        for i, line in enumerate(lines):
            was_valid = line.is_valid
            was_ready = line.is_ready

            if line.did_append_point(p, self.points):
                if line.is_valid:
                    if not was_valid:
                        if self._has_coincident_nested_line(line, nested_lines):
                            # trend already added
                            remove_indexes.append(i)
                            continue

                        # We can't add all trend points because it would form a trend line
                        # on a deeper level and would cause an infinite loop.
                        _insort_point_left(new_trend_points, line.trend_points[0], no_repeat=True)
                        _insort_point_left(new_trend_points, line.trend_points[-1], no_repeat=True)

                elif line.is_ready and not was_ready and self._has_coincident_line(line, lines):
                    # already tracking this trend
                    remove_indexes.append(i)
                    continue

        for i in reversed(remove_indexes):
            del lines[i]

        self._try_fork_line(lines, p, is_upper_trend=is_upper_trend)

        return new_trend_points

    def _try_fork_line(self, lines, p, is_upper_trend=True):
        if len(lines) != 0:
            last_line = lines[-1]
            if not last_line.is_ready or p[0] - last_line.trend_points[0][0] < last_line.min_width:
                return None

        line = self._create_line(is_upper_trend=is_upper_trend)
        line.did_append_point(p, self.points)
        lines.append(line)
        return line

    def _create_line(self, is_upper_trend=True):
        return TrendLine(self.min_width,
            is_upper_trend=is_upper_trend,
            min_points=self.min_points, 
            search_length_rel=self.search_length_rel, 
            search_length_int_rel=self.search_length_int_rel, 
            x_tol_rel=self.x_tol_rel, 
            x_tol_abs=self.x_tol_abs, 
            y_tol_rel=self.y_tol_rel, 
            y_tol_abs=self.y_tol_abs)

    def _has_coincident_nested_line(self, line, nested_lines):
        for lines in nested_lines:
            if self._has_coincident_line(line, lines):
                return True
        return False

    def _has_coincident_line(self, line, lines):
        for other_line in lines:
            if other_line == line or not other_line.is_ready or not line.search_interval.intersects(other_line.search_interval):
                continue
            if other_line.is_line_coincident(line):
                return True
        return False

class TrendLine(Sequence):

    PREDICTION_PHASE_NONE = 0
    PREDICTION_PHASE_INTERSECTED = 1
    PREDICTION_PHASE_RETESTED = 2
    PREDICTION_PHASE_CONFIRMED = 3
    PREDICTION_PHASE_FULFILLED = 4

    @property
    def is_lower_trend(self):
        return not self.is_upper_trend

    @property
    def is_valid(self):
        """Return `True` if all trend line requirements are satisfied."""
        trend_points_count = len(self.trend_points)

        if not self.is_ready or self.width < self.min_width or trend_points_count < self.min_points:
            return False

        # Check distance between trend points
        distance_valid_count = 1
        for i in range(1, trend_points_count):
            if self.trend_points[i][0] - self.trend_points[i - 1][0] >= self.min_point_distance:
                distance_valid_count += 1
                if distance_valid_count >= self.min_points:
                    break

        if distance_valid_count < self.min_points:
            return False

        return True

    @property
    def is_ready(self):
        """Return `True` if there is enough trend points to form a line."""
        return self._line is not None 

    @property
    def is_empty(self):
        """Return `True` where there are no trend points."""
        return len(self.trend_points) == 0

    @property
    def width(self):
        return self.trend_points[-1][0] - self.trend_points[0][0] if len(self.trend_points) >= 2 else 0

    def line_extended(self, width_mult=None, width_max=None, intersection_index=None):
        if not self.is_ready:
            return None
        x0, y0 = self._line[0]
        x1 = self._line[1][0]
        intersection_count = len(self.intersections)
        if intersection_index is not None and intersection_count != 0:
            if intersection_index < 0:
                intersection_index = max(0, len(self.intersections) + intersection_index)
            else:
                intersection_index = min(intersection_index, len(self.intersections) - 1)
            intersection = self.intersections[intersection_index]
            if intersection[0] > x1:
                x1 = intersection[0]
        if width_mult is not None:
            x1 = x0 + (x1 - x0) * width_mult
        if width_max is not None and x1 - x0 > width_max:
            x1 = x0 + width_max
        if intersection_index is not None and intersection_index + 1 < intersection_count:
            next_intersection = self.intersections[intersection_index + 1]
            if next_intersection[0] < x1:
                x1 = next_intersection[0]            
        y1 = self.y(x1)
        return [(x0, y0), (x1, y1)]

    def __init__(self, min_width, is_upper_trend=True, min_points=None, min_point_distance=0, search_length_rel=None, search_length_int_rel=None, x_tol_rel=None, x_tol_abs=None, y_tol_rel=None, y_tol_abs=None):
        self.min_width = min_width
        self.is_upper_trend = is_upper_trend
        self.min_points = min_points
        self.min_point_distance = min_point_distance
        self.search_length_rel = search_length_rel
        self.search_length_int_rel = search_length_int_rel
        self.x_tol_rel = x_tol_rel
        self.x_tol_abs = x_tol_abs
        self.y_tol_rel = y_tol_rel
        self.y_tol_abs = y_tol_abs

        """A list of points through which the trend line is formed."""
        self.trend_points = []
        """An unfiltered list of points from which trend points are derived."""
        self._boundary_points = []
        """A list of points which touch the trend line without the underlying trend crossing the trend."""
        self.tangent_points = []

        self._line = None
        self.search_interval = Interval.empty()
        self.reset_points_of_interest()
        self.reset_prediction()

    def __len__(self):
        return 2 if self.is_ready else 0

    def __getitem__(self, i):
        return self._line[i]

    def __iter__(self):
        if not self.is_ready:
            return
        yield self._line[0]
        yield self._line[1]

    def __copy__(self):
        line = TrendLine(self.min_width,
            is_upper_trend=self.is_upper_trend,
            min_points=self.min_points,
            min_point_distance=self.min_point_distance,
            search_length_rel=self.search_length_rel,
            search_length_int_rel=self.search_length_int_rel,
            x_tol_rel=self.x_tol_rel, 
            x_tol_abs=self.x_tol_abs, 
            y_tol_rel=self.y_tol_rel, 
            y_tol_abs=self.y_tol_abs)
        line.trend_points = list(self.trend_points)
        line._boundary_points = list(self._boundary_points)
        line._line = tuple(self._line)
        line.search_interval = self.search_interval.copy()
        line.intersections = list(self.intersections)
        line.tangent_points = list(self.tangent_points)
        return line

    def copy(self):
        return copy.copy(self)

    def y(self, x):
        """Return the y value at which the `line` crosses the vertical line at `x`."""
        return _line_y(self._line, x)

    def x(self, y):
        """Return the x value at which the `line` crosses the horizontal line at `y`."""
        return _line_x(self._line, y)

    def is_point_on_trend(self, p):
        if self._line is None or not self.search_interval.contains(p[0]):
            return False
        return p[1] >= self._y_min(p[0]) and p[1] <= self._y_max(p[0])
        # return self._is_value_within_tolerance(p[1], self.y(p[0]))
        # return self._is_line_fit_within_tolerance(self.trend_points + [p])

    def is_line_coincident(self, line):
        for p_line in line._line:
            if not self.is_point_on_trend(p_line):
                return False
        return True
        # return self._is_line_fit_within_tolerance(self.trend_points + line.trend_points)

    def set_trend_with_boundary_points(self):
        extremas = list(self._boundary_points)
        coincident_points = []

        # remove all linear and convex (concave) maximas (minimas)
        did_remove = True
        while did_remove:
            did_remove = False
            for i in reversed(range(0, len(extremas) - 2)):
                points = extremas[i:i + 3]
                is_coincident = _is_coincident_points(points, y_tol_rel=self.y_tol_rel, y_tol_abs=self.y_tol_abs)
                is_convex = _is_strictly_convex_points(points, y_tol_rel=self.y_tol_rel, y_tol_abs=self.y_tol_abs)
                if is_coincident or is_convex == self.is_upper_trend:
                    p = extremas[i + 1]
                    if is_coincident and (len(coincident_points) == 0 or coincident_points[-1][0] != p[0]):
                        coincident_points.append(p)
                    if not is_coincident:
                        del self._boundary_points[i + 1]
                    del extremas[i + 1]
                    did_remove = True

        # remove all points which are too close together
        for i in reversed(range(0, len(extremas) - 2)):
            p0, p1, p2 = extremas[i:i + 3]
            if p1[0] - p0[0] < self.min_point_distance or p2[0] - p1[0] < self.min_point_distance:
                del extremas[i + 1]

        final_line = extremas

        extremas_len = len(extremas)
        if extremas_len > 2:
            # Reduce to 2 extremas
            # keep extremas which minimise square error
            lines = []
            for i in range(0, extremas_len - 1):
                lines.append(extremas[i:i + 2])

            def line_fitness(line):
                error_sum = 0
                for p in self._boundary_points:
                    error_sum += math.pow(p[1] - _line_y(line, p[0]), 2.0)
                return error_sum

            lines = sorted(lines, key=line_fitness)
            final_line = lines[0]

        def add_coincident_points(line, coincident_points):
            line_len = len(line)
            if line_len < 2:
                return line
            elif line_len > 2:
                raise Exception("Expected line")
            resulting_points = []
            for p in coincident_points:
                if _is_point_on_line(line, p, y_tol_rel=self.y_tol_rel, y_tol_abs=self.y_tol_abs):
                    resulting_points.append(p)
            resulting_points.insert(0, line[0])
            resulting_points.append(line[1])
            return resulting_points

        final_points = add_coincident_points(final_line, coincident_points)
        self.trend_points = final_points

    def did_append_point(self, p, underlying_points):
        if self.is_ready and not self.search_interval.contains(p[0]):
            return False

        if not self.is_valid:
            _insort_point_left(self._boundary_points, p, no_repeat=True)
            self.update_trend(underlying_points)
            return True
        elif self.is_point_on_trend(p):
            # Point is on the existing line.
            _insort_point_left(self._boundary_points, p, no_repeat=True)
            _insort_point_left(self.trend_points, p, no_repeat=True)
            self.update_trend(underlying_points)
            return True
        else:
            did_change = self.update_points_of_interest(underlying_points)
            self.update_prediction(underlying_points)
            return did_change
        
        return False

    def begin_update(self, domain, underlying_points):
        if not self.search_interval.intersects(domain):
            return

        # remove stale points
        did_remove = False

        def remove_stale(points):
            did_remove = False
            for i in reversed(range(len(points))):
                p = points[i]
                if domain.contains(p[0]):
                    del points[i]
                    did_remove = True
                elif p[0] < domain.start:
                    break
            return did_remove

        if remove_stale(self.trend_points):
            did_remove = True
        if remove_stale(self._boundary_points):
            did_remove = True
        if remove_stale(self.intersections):
            did_remove = True
        if remove_stale(self.tangent_points):
            did_remove = True
        if remove_stale(self.extremas):
            did_remove = True

        if did_remove:
            self.update_trend(underlying_points)

    def update_trend(self, underlying_points):
        if not self.is_valid:
            self.set_trend_with_boundary_points()

        self._line = _line_fit(self.trend_points)
        self._update_trend_tolerance_lines()
        self.reset_points_of_interest()
        self.update_points_of_interest(underlying_points)
        self.update_interval()
        self.reset_prediction()
        self.update_prediction(underlying_points)

    def update_interval(self):
        len_trend_points = len(self.trend_points)
        if self._line is None:
            if len_trend_points == 1:
                self.search_interval = Interval.point(self.trend_points[0][0])
            else:
                self.search_interval = Interval.empty()
            return

        p0, p1 = self._line
        if p0[0] == p1[0]:
            # vertical line
            self.search_interval = Interval.point(p0[0])
            return

        if self.search_length_rel <= 0:
            self.search_interval = Interval.closed(p0[0], p1[0])
            return

        x_min = p0[0]
        x_max = p1[0]

        intersection = self.intersections[0] if len(self.intersections) != 0 else None
        if intersection is not None and intersection[0] <= x_min:
            intersection = None

        if intersection is None or len_trend_points >= self.min_points:
            # search ahead
            x_length = (x_max - x_min) * self.search_length_rel
        else:
            # not enought points to go through intersection
            x_length = intersection[0] - x_min

        # if intersection is None:
        #     # search ahead
        #     x_length = (x_max - x_min) * self.search_length_rel
        # elif len_trend_points >= self.min_points:
        #     # search ahead of intersection
        #     x_length = (intersection[0] - x_min) * self.search_length_int_rel
        # else:
        #     # not enought points to go through intersection
        #     x_length = intersection[0] - x_min

        x_max = x_min + x_length

        self.search_interval = Interval.closed_open(x_min, x_max)

    def reset_points_of_interest(self):
        self.intersections = []
        self.tangent_points = []
        self.extremas = []

    def update_points_of_interest(self, underlying_points):
        if not self.is_ready:
            return False

        i_end = len(underlying_points)
        if i_end == 0:
            return False

        def process_points_on_trend(points_on_trend, insert_index):
            if len(points_on_trend) == 0:
                return
            closest_point = None
            closest_dist = 0
            for p in points_on_trend:
                dist = abs(self.y(p[0]) - p[1])
                if closest_point is None or dist < closest_dist:
                    closest_dist = dist
                    closest_point = p
            self.tangent_points.insert(insert_index, closest_point)

        # find intersections, tangent points and extremas
        phase = 0
        phase_point = None
        previous_phase = phase
        intersection_found = False
        intersection_insert_index = len(self.intersections)
        tangent_insert_index = len(self.tangent_points)
        points_on_trend = []
        new_extremas = []
        extrema = None
        extrema_dist = 0

        # search up to the last intersection or tangent point
        update_interval = self.search_interval
        if len(self.intersections) != 0:
            update_interval = Interval.intersection([update_interval, Interval.gte(self.intersections[-1][0])])
        if len(self.tangent_points) != 0:
            update_interval = Interval.intersection([update_interval, Interval.gt(self.tangent_points[-1][0])])

        for i in reversed(range(len(underlying_points))):
            p = underlying_points[i]
            if not update_interval.contains(p[0]):
                break
            
            if self.is_point_on_trend(p):
                points_on_trend.insert(0, p)
                if extrema is not None:
                    new_extremas.insert(0, extrema)
                    extrema = None
                continue
            elif len(points_on_trend) != 0:
                process_points_on_trend(points_on_trend, tangent_insert_index)
                points_on_trend = []

            phase = p[1] - self.y(p[0])
            at_intersection = previous_phase != 0 and phase * previous_phase < 0

            if at_intersection:
                # found intersection
                intersection = _line_intersection(self._line, (phase_point, p))
                if intersection is None or intersection[0] < self._line[0][0]:
                    intersection = _average_point_of_2(phase_point, p)
                self.intersections.insert(intersection_insert_index, intersection)
                intersection_found = True

                if extrema is not None:
                    new_extremas.insert(0, extrema)
                    extrema = None

            if phase > 0:
                dist = p[1]
            else:
                dist = -p[1]
            if extrema is None or dist > extrema_dist:
                extrema = p
                extrema_dist = dist

            previous_phase = phase
            phase_point = p

        if len(points_on_trend) != 0:
            process_points_on_trend(points_on_trend, tangent_insert_index)

        if extrema is not None:
            new_extremas.insert(0, extrema)
        if len(new_extremas) != 0:
            # merge extremas with existing ones
            if len(self.extremas) == 0:
                self.extremas += new_extremas
            else:
                if self.extremas[-1][0] >= update_interval.start:
                    # old extrema may be outdated
                    del self.extremas[-1]
                self.extremas += new_extremas

        if intersection_found:
            self.update_interval()

        return intersection_found

    def reset_prediction(self):
        self.prediction_phase = TrendLine.PREDICTION_PHASE_NONE
        self.prediction_origin = None
        self.prediction_intersection = None
        self.prediction_extrema = None
        self.prediction_retest_point = None
        self.prediction_retest_x_max = None
        self.prediction_trend_line = None
        self.prediction_confirmation_line = None
        self.prediction_y = None
        self.prediction_y_min = None
        self.prediction_y_max = None
        self.prediction_x_min = None
        self.prediction_x_max = None
        self.prediction_line = None
        self.prediction_failed = False
        self.prediction_succeeded = False
        self.prediction_success_point = None
        self.prediction_valid = False
        self._prediction_last_seen_extrema_point = None
        self._prediction_x_processed = None
        self._prediction_tangent_before_intersection_count = 0

    def update_prediction(self, underlying_points):
        if not self.is_valid:
            return

        self.prediction_origin = self._line[0]

        TYPE_EXTREMA = 1
        TYPE_TANGENT = 2
        TYPE_INTERSECTION = 3

        if self.is_upper_trend:
            extrema_filter = filter(lambda p: p[1] > self.y(p[0]), self.extremas)
        else:
            extrema_filter = filter(lambda p: p[1] < self.y(p[0]), self.extremas)
        extremas = list(map(lambda p: (p[0], p[1], TYPE_EXTREMA), extrema_filter))
        tangents = list(map(lambda p: (p[0], p[1], TYPE_TANGENT), self.tangent_points))
        intersections = list(map(lambda p: (p[0], p[1], TYPE_INTERSECTION), self.intersections))
        points = sorted(extremas + tangents + intersections, key=lambda p: p[0])

        def update_prediction_target():
            nonlocal points

            if self.prediction_intersection is None:
                raise Exception("Should have seen an intersection: {}".format(points))
            if self._prediction_last_seen_extrema_point is None:
                return False

            self.prediction_extrema = self._prediction_last_seen_extrema_point
            self.prediction_confirmation_line = [self.prediction_origin, self.prediction_extrema]
            self.prediction_y = self.prediction_intersection[1] + (self.prediction_extrema[1] - self.prediction_intersection[1]) * GOLD_2
            self.prediction_y_min = self._value_tolerance_min(self.prediction_y)
            self.prediction_y_max = self._value_tolerance_max(self.prediction_y)
            self.prediction_x_min = self.prediction_extrema[0]
            self.prediction_x_max = self.prediction_x_min + (self.prediction_extrema[0] - self.prediction_origin[0]) * GOLD
            self.prediction_line = [(self.prediction_extrema[0], self.prediction_y), (self.prediction_x_max, self.prediction_y)]
            self.prediction_trend_line = [self.prediction_origin, (self.prediction_x_max, self.y(self.prediction_x_max))]

            self.prediction_valid = self.prediction_y > 0

            return True

        if len(points) == 0:
            return

        for p in points:
            if self._prediction_x_processed is not None and p[0] < self._prediction_x_processed:
                continue
            point_type = p[2]
            clean_point = (p[0], p[1])
            if self.prediction_succeeded or self.prediction_failed:
                break
            elif self.prediction_phase == TrendLine.PREDICTION_PHASE_NONE:
                # Look for intersection
                if point_type == TYPE_TANGENT:
                    self._prediction_tangent_before_intersection_count += 1
                if point_type == TYPE_INTERSECTION:
                    if self._prediction_tangent_before_intersection_count < self.min_points:
                        self.prediction_failed = True
                        self.prediction_valid = False
                        break
                    self.prediction_phase = TrendLine.PREDICTION_PHASE_INTERSECTED
                    self.prediction_intersection = clean_point
                    self.prediction_retest_x_max = self.prediction_intersection[0] + (self.prediction_intersection[0] - self.prediction_origin[0]) / GOLD
                    self.prediction_trend_line = [self.prediction_origin, (self.prediction_retest_x_max, self.y(self.prediction_retest_x_max))]
                continue

            # Look for extrema and retest points
            if point_type == TYPE_EXTREMA:
                self._prediction_last_seen_extrema_point = clean_point

                if self.prediction_phase == TrendLine.PREDICTION_PHASE_INTERSECTED:
                    # Update prediction targets
                    if not update_prediction_target():
                        self.prediction_failed = True
                        break

                if self.prediction_phase == TrendLine.PREDICTION_PHASE_RETESTED:
                    # Check for confirmation
                    confirmation_y = _line_y(self.prediction_confirmation_line, p[0])
                    confirmation_y_min = self._value_tolerance_min(confirmation_y)
                    confirmation_y_max = self._value_tolerance_max(confirmation_y)
                    if (self.is_upper_trend and p[1] >= confirmation_y_max) or (self.is_lower_trend and p[1] <= confirmation_y_min):
                        self.prediction_phase = TrendLine.PREDICTION_PHASE_CONFIRMED
                
                if self.prediction_phase == TrendLine.PREDICTION_PHASE_CONFIRMED:
                    # Check for prediction fulfilment
                    if (self.is_upper_trend and p[1] >= self.prediction_y_min) or (self.is_lower_trend and p[1] <= self.prediction_y_max):
                        self.prediction_phase = TrendLine.PREDICTION_PHASE_FULFILLED
                        self.prediction_succeeded = True
                        self.prediction_success_point = clean_point
                        break
                    elif p[0] >= self.prediction_x_max:
                        self.prediction_failed = True
                        break
            elif point_type == TYPE_TANGENT:
                if p[0] > self.prediction_retest_x_max:
                    # Retest too far from intersection
                    self.prediction_failed = True
                    break
                # Retested, prediction is ready.
                # Note: There may be more than one retest, take the last one
                if not update_prediction_target():
                    self.prediction_failed = True
                    break
                self.prediction_phase = TrendLine.PREDICTION_PHASE_RETESTED
                self.prediction_retest_point = clean_point
            elif point_type == TYPE_INTERSECTION:
                # Intersected the trend line again, i.e. retest failed
                self.prediction_failed = True
                break

        if not self.prediction_succeeded and not self.prediction_failed:
            self._prediction_x_processed = points[-1][0]

            # Check prediction timeouts
            last_underlying_point = underlying_points[-1]
            if self.prediction_phase >= TrendLine.PREDICTION_PHASE_RETESTED:
                if last_underlying_point[0] >= self.prediction_x_max:
                    self.prediction_failed = True
            if self.prediction_phase == TrendLine.PREDICTION_PHASE_INTERSECTED:
                if last_underlying_point[0] >= self.prediction_retest_x_max:
                    self.prediction_failed = True

    def _is_line_fit_within_tolerance(self, points):
        if len(points) <= 2:
            return True

        fit_f = _line_fit_f(points)
        for p in points:
            y_line = fit_f(p[0])
            if not self._is_value_within_tolerance(p[1], y_line):
                return False
        return True

    def _is_value_within_tolerance(self, y, ref_y):
        if y is None or ref_y is None:
            return False
        return y < self._value_tolerance_min(ref_y) or y > self._value_tolerance_max(ref_y)

    def _value_tolerance_min(self, y):
        y_min = None
        if self.y_tol_abs is not None:
            y_min = y - self.y_tol_abs
        if self.y_tol_rel is not None and abs(y) > EPSILON_ZERO:
            y_min_rel = y * (1 - self.y_tol_rel)
            if y_min is None:
                y_min = y_min_rel
            else:
                y_min = max(y_min, y_min_rel)
        if y_min is None:
            y_min = y
        return y_min

    def _value_tolerance_max(self, y):
        y_max = None
        if self.y_tol_abs is not None:
            y_max = y + self.y_tol_abs
        if self.y_tol_rel is not None and abs(y) > EPSILON_ZERO:
            y_max_rel = y * (1 + self.y_tol_rel)
            if y_max is None:
                y_max = y_max_rel
            else:
                y_max = min(y_max, y_max_rel)
        if y_max is None:
            y_max = y
        return y_max
    
    def _y_min(self, x):
        return _line_y(self._tolerance_line_lower, x)

    def _y_max(self, x):
        return _line_y(self._tolerance_line_upper, x)

    def _update_trend_tolerance_lines(self):
        self._tolerance_line_upper = self._calculate_trend_tolerance_line(is_upper=True)
        self._tolerance_line_lower = self._calculate_trend_tolerance_line(is_upper=False)

    def _calculate_trend_tolerance_line(self, is_upper=True):
        line = self._line

        trend_points_count = len(self.trend_points)
        if trend_points_count < 2:
            return None
        # elif trend_points_count == 2:
        else:
            x0 = line[0][0]
            x1 = line[1][0]
            if is_upper:
                y0 = self._value_tolerance_max(line[0][1])
                y1 = self._value_tolerance_max(line[1][1])
            else:
                y0 = self._value_tolerance_min(line[0][1])
                y1 = self._value_tolerance_min(line[1][1])
            return [(x0, y0), (x1, y1)]

        # slope = (line[1][1] - line[0][1]) / (line[1][0] - line[0][0])
        # const = _line_y(line, 0)
        # params = [const, slope]

        # def error_func(params_attempt):
        #     const, slope = params_attempt
        #     error_sum = 0

        #     for p in self.trend_points:
        #         y_line = const + slope * p[0]
        #         if is_upper:
        #             y_max = self._value_tolerance_max(p[1])
        #             err = (y_line - y_max) ** 2
        #             if y_line > y_max:
        #                 err *= 1e5
        #         else:
        #             y_min = self._value_tolerance_min(p[1])
        #             err = (y_line - y_min) ** 2
        #             if y_line < y_min:
        #                 err *= 1e5
        #         error_sum += err
        #     return error_sum
            

        # # docs: https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize
        # x0 = list(params)
        # method = 'Nelder-Mead'
        # tolerance = 1e-6
        # result = minimize(error_func, x0, method=method, tol=tolerance)
        # params = list(result.x)
        # const, slope = params
        # x0 = line[0][0]
        # y0 = const + slope * x0
        # x1 = line[1][0]
        # y1 = const + slope * x1
        # return [(x0, y0), (x1, y1)]

def _line_fit_f(points):
    points_t = np.transpose(points)
    x = points_t[0]
    y = points_t[1]
    fit = np.polyfit(x, y, 1)
    fit_f = np.poly1d(fit)
    return fit_f

def _line_fit(points):
    if len(points) < 2:
        return None
    if len(points) == 2:
        return points

    # describe line with two points
    fit_f = _line_fit_f(points)
    x0 = points[0][0]
    x1 = points[-1][0]
    p1 = [x0, fit_f(x0)]
    p2 = [x1, fit_f(x1)]

    return [p1, p2]

def _is_line_fit_within_tolerance(points, y_tol_rel=None, y_tol_abs=None):
    if len(points) <= 2:
        return True

    fit_f = _line_fit_f(points)
    for p in points:
        y_line = fit_f(p[0])
        if not _is_value_within_tolerance(p[1], y_line, y_tol_rel=y_tol_rel, y_tol_abs=y_tol_abs):
            return False
    return True

def _line_intersection(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
       return None

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return [x, y]

def _is_strictly_convex_points(points, y_tol_rel=None, y_tol_abs=None):
    """
    A function is strictly convex if the line segment between any two points on the graph of the function lies above the graph.
    """
    if len(points) != 3:
        raise Exception("Expected 3 points")
    x = points[1][0]
    y = points[1][1]
    y0 = _line_y((points[0], points[2]), x)
    
    if y0 is None:
        return False
    if y_tol_abs is not None and y0 - y < y_tol_abs:
        return False
    if y_tol_rel is not None and abs(y0) > EPSILON_ZERO and abs(y) > EPSILON_ZERO and y0 / y - 1 < y_tol_rel:
        return False
    return True

def _is_coincident_points(points, y_tol_rel=None, y_tol_abs=None):
    if len(points) != 3:
        raise Exception("Expected 3 points")
    return _is_point_on_line((points[0], points[2]), points[1], y_tol_rel=y_tol_rel, y_tol_abs=y_tol_abs)

def _is_point_on_line(line, p, y_tol_rel=None, y_tol_abs=None):
    x = p[0]
    y = p[1]
    y0 = _line_y(line, x)
    return _is_value_within_tolerance(y, y0, y_tol_rel=y_tol_rel, y_tol_abs=y_tol_abs)

def _is_value_within_tolerance(y, y0, y_tol_rel=None, y_tol_abs=None):
    if y is None or y0 is None:
        return False
    if y_tol_abs is not None and abs(y - y0) > y_tol_abs:
        return False
    if y_tol_rel is not None and abs(y0) > EPSILON_ZERO and abs(y) > EPSILON_ZERO and abs(y / y0 - 1) > y_tol_rel:
        return False
    return True

def _line_y(line, x):
    """Return the y value at which the `line` crosses the vertical line at `x`."""
    p1 = line[0]
    p2 = line[1]
    if p2[0] == p1[0]:
        if p1[0] == x:
            return p1[1]
        return None
    m = (p2[1] - p1[1]) / (p2[0] - p1[0])
    y = p1[1] + m * (x - p1[0])
    return y

def _line_x(line, y):
    """Return the x value at which the `line` crosses the horizontal line at `y`."""
    p1 = line[0]
    p2 = line[1]
    if p2[0] == p1[0]:
        return p1[0]
    m = (p2[1] - p1[1]) / (p2[0] - p1[0])
    if m == 0:
        if p1[1] == y:
            return p1[0]
        return None
    x = p1[0] + (y - p1[1]) / m
    return x

def _insort_points_left(sorted_points, insert_points, lo=0, hi=None, no_repeat=False):
    for p in insert_points:
        _insort_point_left(sorted_points, p, lo=lo, hi=hi, no_repeat=no_repeat) 

def _insort_point_left(a, p, lo=0, hi=None, no_repeat=False):
    """
    Insert point `p` in list `a`, and keep it sorted assuming `a` is sorted.
    If `p` is already in `a`, insert it to the left of the leftmost `p` if `no_repeat` is `False`.
    Optional args `lo` (default `0`) and `hi` (default `len(a)`) bound the
    slice of `a` to be searched.

    Source: https://github.com/python/cpython/blob/master/Lib/bisect.py
    """

    if lo < 0:
        raise ValueError('lo must be non-negative')
    a_len = len(a)
    if hi is None:
        hi = a_len
    while lo < hi:
        mid = (lo + hi) // 2
        if a[mid][0] < p[0]: lo = mid + 1
        else: hi = mid
    if no_repeat and lo < a_len and p[0] == a[lo][0]:
        # repeated point
        return
    a.insert(lo, p)

def _average_point_of_2(p1, p2):
    return ((p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2)


from .extension import Extension
from ..constant import Constant
from intervalpy import Interval

class ConstantExtension(Extension):

    """
    Extends an end of a function with a line using its edge values.
    """

    name = "constant"

    def update_extension(self):
        if self.curve.domain.is_empty:
            self.start_func.value = None
            self.end_func.value = None
            return

        if self.start:
            x = self.curve.domain.start
            y = self.curve.y(x)
            self.start_func.value = y

        if self.end:
            x = self.curve.domain.end
            y = self.curve.y(x)
            self.end_func.value = y

    def create_extension_func(self, start=False):
        return Constant(0)



from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Generic(Curve):

    def get_domain(self):
        return self._domain

    def __init__(self, y_func, domain=None, min_step=MIN_STEP):
        super().__init__(min_step=min_step)
        self.y_func = y_func
        self._domain = Interval.parse(domain) if domain is not None else Interval.infinite()

    def y(self, x):
        if not self.domain.contains(x):
            return None
        return self.y_func(x)


from .curve import Curve, MIN_STEP
from .map import Map
from .points import Points
from intervalpy import Interval
from pyduration import Duration

OHLC_KEYS = ['open', 'high', 'low', 'close', 'volume']
OHLC_POINT_COUNT = len(OHLC_KEYS)
OHLC_STEPS = OHLC_POINT_COUNT - 1
QUOTE_MIN_STEP_COEFFICIENT = 0.1

OPEN = 0
HIGH = 1
LOW = 2
CLOSE = 3
VOLUME = 4


class Quotes(Curve):
    """
    Represents quotes as [open, high, low, close, (volume)].
    """

    @property
    def close(self):
        if self._close is None:
            self._close = Map(
                self,
                lambda q: q[CLOSE],
                skip_none=True,
                name='close'
            )
        return self._close

    @property
    def open(self):
        if self._open is None:
            self._open = Map(
                self,
                lambda q: q[OPEN],
                skip_none=True,
                name='open'
            )
        return self._open

    @property
    def high(self):
        if self._high is None:
            self._high = Map(
                self,
                lambda q: q[HIGH],
                skip_none=True,
                name='high'
            )
        return self._high

    @property
    def low(self):
        if self._low is None:
            self._low = Map(
                self,
                lambda q: q[LOW],
                skip_none=True,
                name='low'
            )
        return self._low

    @property
    def hl2(self):
        if self._hl2 is None:
            self._hl2 = Map(
                self,
                lambda q: (q[HIGH] + q[LOW]) / 2,
                skip_none=True,
                name='hl2'
            )
        return self._hl2

    @property
    def volume(self):
        if self._volume is None:
            self._volume = Map(
                self,
                lambda q: q[VOLUME],
                skip_none=True,
                name='volume'
            )
        return self._volume

    def get_domain(self):
        return self._quote_points.domain
        # if domain.is_empty:
        #     return domain
        # return self.duration.span(domain, start_open=False)

    def __init__(self, duration, quote_points=None, **kwargs):
        """
        `quote_points` are assumed to be in the form [timestamp, [o, h, l, c, v]]
        in strict ascending order with no gaps and with the same duration as the receiver.
        """
        duration = Duration.parse(duration)
        self.duration = duration
        super().__init__(min_step=duration.min_seconds * 0.01, **kwargs)
        self._quote_points = Points(
            [], interpolation=Points.interpolation.previous, uniform=duration.is_uniform)
        self._quote_points.add_observer(begin=self.begin_update, end=self.end_update, prioritize=True)
        self._close = None
        self._open = None
        self._high = None
        self._low = None
        self._hl2 = None
        self._ohlc = None
        self._volume = None
        if quote_points is not None:
            self.set(quote_points)

    def __repr__(self):
        try:
            return f'quotes({self.duration})'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def last_quote(self):
        return self.y_end()

    def first_quote(self):
        return self.y_start()

    def get_range(self, domain=None, **kwargs):
        quotes = self.sample(domain=domain, **kwargs)
        low = None
        high = None
        for q in quotes:
            if low is None or q.low < low:
                low = q.low
            if high is None or q.high > high:
                high = q.high
        if low is None or high is None:
            return Interval.empty()
        return Interval(low, high)

    def ao(self):
        """
        The Awesome Oscillator is an indicator used to measure market momentum.
        AO calculates the difference of a 34 Period and 5 Period Simple Moving Averages.
        The Simple Moving Averages that are used are not calculated using closing
        price but rather each bar's midpoints. AO is generally used to affirm trends
        or to anticipate possible reversals.
        """
        uniform = self.duration.is_uniform
        return self.hl2.sma(5, uniform=uniform) - self.hl2.sma(34, uniform=uniform)

    def alligator(self):
        """
        The Alligator indicator uses three smoothed moving averages, set at five,
        eight and 13 periods, which are all Fibonacci numbers. The initial smoothed
        average is calculated with a simple moving average (SMA), adding additional
        smoothed averages that slow down indicator turns.
        """
        uniform = self.duration.is_uniform
        jaw = self.hl2.sma(13, uniform=uniform).offset(8, duration=self.duration)
        teeth = self.hl2.sma(8, uniform=uniform).offset(5, duration=self.duration)
        lips = self.hl2.sma(5, uniform=uniform).offset(3, duration=self.duration)
        return jaw, teeth, lips

    def trailing_high(self, degree, is_period=False):
        return self.high.trailing_max(degree, is_period=is_period, interpolation=-1, uniform=self.duration.is_uniform)

    def trailing_low(self, degree, is_period=False):
        return self.low.trailing_max(degree, is_period=is_period, interpolation=-1, uniform=self.duration.is_uniform)

    def append(self, quote_point):
        """
        `quote_points` are assumed to be in the form [timestamp, [o, h, l, c, v]]
        in strict ascending order with no gaps and with the same duration as the receiver.
        """
        self.append_list([quote_point])

    def append_list(self, quote_points):
        """
        `quote_points` are assumed to be in the form [timestamp, [o, h, l, c, v]]
        in strict ascending order with no gaps and with the same duration as the receiver.
        """
        if self.domain.is_empty:
            return self.set(quote_points)
        self._quote_points.append_list(quote_points)

    def set(self, quote_points):
        """
        `quote_points` are assumed to be in the form [timestamp, [o, h, l, c, v]]
        in strict ascending order with no gaps and with the same duration as the receiver.
        """
        self._quote_points.set(quote_points)

    def sample(self, domain=None, min_step=MIN_STEP, step=None):
        points = self.sample_points(
            domain=domain,
            min_step=min_step,
            step=step
        )
        return [p[1] for p in points]

    def sample_points(self, domain=None, min_step=MIN_STEP, step=None):
        domain = Interval.parse(domain, default_inf=True)
        if domain.is_empty:
            return []
        pinterval = self.duration.span(domain, start_open=False)
        return self._quote_points.sample_points(domain=pinterval, min_step=min_step, step=step)

    def y(self, x):
        return self._quote_points.y(x)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        # return self.duration.next(x)
        return self._quote_points.x_next(x, min_step=min_step, limit=limit)

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        # return self.duration.previous(x)
        return self._quote_points.x_previous(x, min_step=min_step, limit=limit)



import math
from .tangent import TangentExtension
from ..sin import Sin

class SinExtension(TangentExtension):

    """
    Extends an end of a function with a sine wave using its edge tangents.
    """

    name = "sin"

    def __init__(self, func, period=0, **kwargs):
        self.period = float(period)
        assert self.period > 0
        super().__init__(func, **kwargs)

    def create_extension_func(self, start=False):
        return Sin(amplitude=0, period=self.period)

    def update_extension_func(self, sin, x, y, dy):
        # angle
        x_coef = 2 * math.pi / self.period
        angle = x_coef * x

        # phase
        num = x_coef * y * math.cos(angle) - dy * math.sin(angle)
        den = x_coef * y * math.sin(angle) + dy * math.cos(angle)
        phase = math.atan2(num, den)

        # amplitude
        if (angle + phase) % math.pi != 0:
            amplitude = y / math.sin(angle + phase)
        else:
            amplitude = dy / (x_coef * math.cos(angle + phase))

        sin.set(amplitude=amplitude, period=self.period, phase=phase)



from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Piecewise(Curve):

    def get_domain(self):
        return Interval.union(self.domains)

    def __init__(self, funcs, domains):
        """
        `domains` are assumed to be in ascending order and not overlapping.
        """
        super().__init__()
        if len(funcs) != len(domains):
            raise Exception("Must specify sub-domains for all sub-functions in a piecewise function")
        self.funcs = Curve.parse_many(funcs)
        self.domains = Interval.parse_many(domains)
        unique_funcs = list(set(list(self.funcs)))
        self._observer_tokens = list(map(lambda f: f.add_observer(begin=self.begin_update, end=self.end_update, prioritize=True), unique_funcs))

    def __del__(self):
        for func, token in zip(self.funcs, self._observer_tokens):
            func.remove_observer(token)

    def y(self, x):
        for i in range(len(self.domains)):
            d = self.domains[i]
            if not d.contains(x, enforce_start=False):
                continue
            elif not d.contains(x, enforce_end=False):
                break
            f = self.funcs[i]
            return f.y(x)
        return None
    
    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        # FIXME: infinite funcs don't work (unit tests disabled)
        min_step = self.resolve_min_step(min_step)
        if not self.domain.contains(x - min_step, enforce_end=False):
            return None
        x_map = map(lambda f: f.x_previous(x, min_step=min_step, limit=limit), self.funcs)
        x_map_d = zip(x_map, self.domains)
        def x_previous_f(z):
            x, d = z
            return d.contains(x)
        z = max(filter(x_previous_f, x_map_d), default=None)
        x1 = z[0] if z is not None else None
        if x1 is None and self.domains[0].contains(x):
            return self.domains[0].start

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        # FIXME: infinite funcs don't work (unit tests disabled)
        min_step = self.resolve_min_step(min_step)
        if not self.domain.contains(x + min_step, enforce_end=False):
            return None
        x_map = map(lambda f: f.x_next(x, min_step=min_step, limit=limit), self.funcs)
        x_map_d = zip(x_map, self.domains)
        def x_next_f(z):
            x, d = z
            return d.contains(x)
        z = min(filter(x_next_f, x_map_d), default=None)
        x1 = z[0] if z is not None else None
        if x1 is None and self.domains[-1].contains(x):
            return self.domains[-1].end



from .accumulator import Accumulator
from .line import Line
from .points import Points
from .curve import MIN_STEP
from intervalpy import Interval


class Integral(Accumulator):

    def __init__(self, func, const=0, interpolation=None, **kwargs):
        super().__init__(func, self._integral_scan, interpolation=None, **kwargs)
        self.const = const
        self.integral_interpolation = interpolation or Points.interpolation.linear
        self._line = Line(const=0, slope=0)

    def scanned_y(self, x):
        y = super().scanned_y(x)
        if y is None:
            return y
        return y + self.const

    def _integral_scan(self, x, y, integral_sum):
        if y is None:
            return integral_sum
        if integral_sum is None:
            integral_sum = 0.0

        x0 = self.accumulated_points.x_previous(x)
        if x0 is not None:
            y0 = self.curve.y(x0)
        else:
            y0 = None

        if y0 is not None:
            interpolation = self.integral_interpolation
            if interpolation == Points.interpolation.linear:
                p1 = (x0, y0)
                p2 = (x, y)
            elif interpolation == Points.interpolation.previous:
                p1 = (x0, y0)
                p2 = (x, y0)
            elif interpolation == Points.interpolation.next:
                p1 = (x0, y)
                p2 = (x, y)
            else:
                raise Exception(f'Unsupported interpolation: {self.interpolation}')
            self._line.set(p1=p1, p2=p2)
            area = self._line.partial_integration(x0, x)
            integral_sum += area

        return integral_sum



from .extension import Extension
from ..line import Line
from intervalpy import Interval

class TangentExtension(Extension):

    """
    Extends an end of a function with a line using its edge tangents.
    """

    name = "tangent"

    def __init__(self, func, regression_degree=None, regression_period=None, **kwargs):
        self.regression_degree = None
        self.regression_period = None
        if regression_degree is not None:
            self.regression_degree = int(regression_degree)
            assert self.regression_degree > 0
        if regression_period is not None:
            self.regression_period = float(regression_period)
            assert self.regression_period > 0
        assert self.regression_degree is None or self.regression_period is None
        super().__init__(func, **kwargs)

    def update_extension(self):
        if self.start:
            x = self.curve.domain.start
            if x is not None and self.curve.domain.start_open and self.curve.domain.contains(x + self.min_step):
                x += self.min_step
            y = None
            d_y = None
            if self.regression_degree is not None:
                x1 = x
                for i in range(self.regression_degree):
                    x1 = self.curve.x_next(x1, min_step=self.min_step)
                domain = Interval(x, x1)
                tangent = self.curve.regression(domain, min_step=self.min_step)
                if tangent is not None:
                    y = tangent.y(x)
                    d_y = tangent.slope
            elif self.regression_period is not None:
                domain = Interval(x, x + self.regression_period)
                tangent = self.curve.regression(domain, min_step=self.min_step)
                if tangent is not None:
                    y = tangent.y(x)
                    d_y = tangent.slope
            else:
                y = self.curve.y(x)
                d_y = self.curve.d_y(x, forward=True)

            self.start_valid = y is not None and d_y is not None
            if self.start_valid:
                self.update_extension_func(self.start_func, x, y, d_y)

        if self.end:
            x = self.curve.domain.end
            if x is not None and self.curve.domain.end_open and self.curve.domain.contains(x - self.min_step):
                x -= self.min_step
            y = None
            d_y = None
            if self.regression_degree is not None:
                x0 = x
                for i in range(self.regression_degree):
                    x0 = self.curve.x_previous(x0, min_step=self.min_step)
                domain = Interval(x0, x)
                tangent = self.curve.regression(domain, min_step=self.min_step)
                if tangent is not None:
                    y = tangent.y(x)
                    d_y = tangent.slope
            elif self.regression_period is not None:
                domain = Interval(x - self.regression_period, x)
                tangent = self.curve.regression(domain, min_step=self.min_step)
                if tangent is not None:
                    y = tangent.y(x)
                    d_y = tangent.slope
            else:
                y = self.curve.y(x)
                d_y = self.curve.d_y(x, forward=False)

            self.end_valid = y is not None and d_y is not None
            if self.end_valid:
                self.update_extension_func(self.end_func, x, y, d_y)

    def create_extension_func(self, start=False):
        return Line(const=0, slope=0)

    def update_extension_func(self, func, x, y, dy):
        func.set(p1=(x, y), slope=dy)



import inspect
from typing import Mapping, Iterable


def flatten(items) -> list:
    return list(_flatten(items))


def _flatten(items):
    if items is None or isinstance(items, (str, bytes, Mapping)) or not isinstance(items, Iterable):
        yield items
        return
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in _flatten(x):
                yield sub_x
        else:
            yield x


def extend(primary: Mapping, *others: Mapping, in_place=False):
    """
    Copies values from `others` to `primary`.
    """
    others = flatten(others)
    if not in_place:
        primary = dict(primary or {})
    for other in others:
        if other is None:
            continue
        for key, value in other.items():
            primary[key] = value
    return primary


def count_positional_args(f, default=1):
    if not callable(f):
        raise Exception('Expected callable function')
    if inspect.isbuiltin(f):
        return default
    sig = inspect.signature(f)
    count = 0
    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.POSITIONAL_ONLY or \
            param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            count += 1
    return count



import math
from .curve import Curve, MIN_STEP
from .constant import Constant
from intervalpy import Interval

class Scan(Curve):

    def get_domain(self):
        return self.curve.domain

    def __init__(self, func, tfm, min_step=MIN_STEP):
        super().__init__(min_step=min_step)
        self.curve = Curve.parse(func)
        self.tfm = tfm
        self.scan_start = None
        self.current = None
        self._observer_token = self.curve.add_observer(begin=self.begin_scan_update, end=self.end_scan_update, prioritize=True)

    def __del__(self):
        try:
            self.curve.remove_observer(self._observer_token)
        except Exception:
            pass

    def y(self, x):
        if not self.curve.domain.contains(x) and not self.domain.contains(x):
            return None
        self.scan(x)
        return self.scanned_y(x)

    def scanned_y(self, x):
        raise Exception("Not implemented")

    def continue_scan(self, x):
        return self.current < x

    def init_scan(self, x):
        if self.curve.domain.is_empty or self.curve.domain.is_negative_infinite:
            return x
        x0 = self.curve.domain.start
        if not self.domain.contains(x0):
            x0 = self.curve.x_next(x0, min_step=self.min_step)
        return x0

    def offset_scan(self, x):
        return x

    def reset_scan(self):
        self.scan_start = None
        self.current = None

    def scan(self, x0):
        x = self.offset_scan(x0)
        if self.scan_start is not None and x < self.scan_start:
            self.reset_scan()
        while self.current is None or self.continue_scan(x0):
            if self.current is None:
                self.scan_start = self.init_scan(x)
                current = self.scan_start
            else:
                current = self.x_next(self.current, min_step=self.min_step, limit=x0)
            if current is None or not self.curve.domain.contains(current):
                break
            if self.current is not None and current <= self.current:
                raise Exception('Next scan x value ({}) is smaller than or equal to the current scan x value ({}).'.format(current, self.current))
            self.tfm(current, self.curve.y(current))
            self.current = current

    def sample_points(self, domain=None, min_step=MIN_STEP, step=None):
        min_step = self.resolve_min_step(min_step)
        if domain is None:
            domain = self.domain
        else:
            domain = Interval.intersection([self.domain, domain])
        if domain.is_empty:
            return []
        elif domain.is_infinite:
            raise Exception("Cannot sample points on an infinite domain. Specify a finite domain.")
        self.scan(domain.start)
        self.scan(domain.end)
        return super().sample_points(domain=domain, min_step=min_step, step=step)

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.curve.x_previous(x, min_step=min_step, limit=limit)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.curve.x_next(x, min_step=min_step, limit=limit)

    def begin_scan_reset(self):
        self.reset_scan()
        self.begin_update(Interval.infinite())

    def end_scan_reset(self):
        self.end_update(Interval.infinite())

    def begin_scan_update(self, domain):
        domain = self._update_interval(domain)
        if self.current is not None:
            if domain.start < self.current or (domain.start == self.current and not domain.start_open):
                self.reset_scan()

        if self.current is not None and self.scan_start is not None and self.current < self.scan_start:
            self.reset_scan()

        self.begin_update(domain)

    def end_scan_update(self, domain):
        domain = self._update_interval(domain)
        self.end_update(domain)

    def _update_interval(self, domain):
        return Interval(domain.start, math.inf, start_open=domain.start_open, end_open=True)



import math
from .curve import Curve, MIN_STEP
from intervalpy import Interval

PREVIOUS_INTERPOLATION = -1
LINEAR_INTERPOLATION = 0
NEXT_INTERPOLATION = 1

ALL_INTERPOLATIONS = [
    PREVIOUS_INTERPOLATION,
    LINEAR_INTERPOLATION,
    NEXT_INTERPOLATION
]


class Points(Curve):

    class interpolation:

        previous = PREVIOUS_INTERPOLATION
        linear = LINEAR_INTERPOLATION
        next = NEXT_INTERPOLATION

        default = linear
        all = ALL_INTERPOLATIONS

    @property
    def is_uniform(self):
        if self._force_equally_spaced:
            return True
        return self._is_equally_spaced

    def get_domain(self):
        if len(self._points) == 0:
            return Interval.empty()
        return Interval(self._points[0][0], self._points[-1][0], start_open=False, end_open=False)

    def __init__(self, points, interpolation=None, uniform=True):
        """
        `points` are assumed to be strictly ordered in ascending order w.r.t. to `x`.
        """
        super().__init__()
        
        if interpolation is None:
            interpolation = Points.interpolation.default
        if interpolation not in ALL_INTERPOLATIONS:
            raise Exception('Invalid interpolation')
        self.interpolation = interpolation
        self.interval = None
        self._points = []
        self._force_equally_spaced = uniform
        self._is_equally_spaced = None
        self.set(points)

    def append(self, point):
        """
        `points` are assumed to be strictly ordered in ascending order w.r.t. to `x`.
        """
        self.append_list([point])

    def append_list(self, points):
        """
        `points` are assumed to be strictly ordered in ascending order w.r.t. to `x`.
        """
        points_len = len(self._points)
        if points_len == 0:
            return self.set(points)
        new_points_len = len(points)
        if new_points_len == 0:
            return

        if points[0][0] <= self._points[-1][0]:
            raise Exception('Attempting to append points in non-ascending order')

        if self.domain.is_empty:
            domain = Interval.closed(points[0][0], points[-1][0])
        else:
            domain = Interval(self.domain.end, points[-1][0], start_open=True, end_open=False)
        self.begin_update(domain)
        self._is_equally_spaced = self._points_equally_spaced(self._points, points)
        if self._force_equally_spaced and not self._is_equally_spaced:
            raise Exception('Attempting to append points at non-regular intervals: {}{} + {}{}'.format('...' if len(self._points) > 2 else '', self._points[len(self._points) - 2:], points[:2], '...' if len(points) > 2 else ''))
        self._points += points
        self._did_change_points()
        self.end_update(domain)

    def replace(self, point, or_append=False):
        if not self.domain.contains(point[0]):
            if or_append:
                return self.append(point)
            else:
                raise Exception('Attempting to replace point outside of bounds')

        update_domain = Interval.point(point[0])
        self.begin_update(update_domain)

        nearest_i = int(math.ceil(self.x_index(point[0])))
        nearest_p = self._points[nearest_i]
        if point[0] != nearest_p[0]:
            if self._force_equally_spaced:
                raise Exception('Attempting to replace point {} between existing points. The nearest is {}.'.format(point, nearest_p))
            else:
                self._points.insert(nearest_i, point)
                self._is_equally_spaced = False
        else:
            self._points[nearest_i] = point

        self._did_change_points()
        self.end_update(update_domain)    

    def reset(self, domain=None):
        points_len = len(self._points)
        if points_len == 0:
            return
        if domain is None:
            return self.set([])
        domain = Interval.parse(domain)

        if not self.domain.intersects(domain):
            return

        if domain.is_superset_of(self.domain):
            self.set([])
            return

        remove_start_i, remove_end_i = self._domain_indexes(domain)
        if remove_start_i == remove_end_i:
            # Nothing to remove
            return
        head = self._points[:remove_start_i]
        tail = self._points[remove_end_i:]
        if len(head) != 0 and domain.contains(head[-1][0]):
            del head[-1]
        if len(tail) != 0 and domain.contains(tail[0][0]):
            del tail[0]
        head_len = len(head)
        tail_len = len(tail)
        if head_len + tail_len == 0:
            self.set([])
            return
        points = head + tail

        update_start = self._points[0][0]
        update_start_open = False
        if head_len != 0:
            update_start = head[-1][0]
            update_start_open = True

        update_end = self._points[-1][0]
        update_end_open = False
        if tail_len != 0:
            update_end = tail[0][0]
            update_end_open = True

        update_domain = Interval(update_start, update_end, start_open=update_start_open, end_open=update_end_open)

        self.set(points, update_domain=update_domain)

    def set(self, points, update_domain=None):
        """
        `points` are assumed to be strictly ordered in ascending order w.r.t. to `x`.
        """
        if update_domain is None:
            points_len = len(points)
            if points_len == 0:
                update_domain = self.domain
            else:
                update_domain = Interval.union([self.domain, Interval.closed(points[0][0], points[-1][0])])
        self.begin_update(update_domain)
        self._is_equally_spaced = self._points_equally_spaced([], points)
        if self._force_equally_spaced and not self._is_equally_spaced:
            raise Exception('Attempting to set points at non-regular intervals')
        self._points = list(points)
        self._did_change_points()
        self.end_update(update_domain)

    def sample_points(self, domain=None, min_step=None, step=None):
        domain = Interval.parse(domain, default_inf=True)

        if self.domain.is_empty:
            return []

        if self.interval is not None and ((min_step is not None and min_step > self.interval) or (step is not None and step != self.interval)):
            # Irregular sampling
            return super().sample_points(domain=domain, min_step=min_step, step=step)

        if domain is None or domain.is_superset_of(self.domain):
            # Sample all
            return list(self._points)

        # Sample some
        domain = Interval.intersection([self.domain, domain])
        if domain.is_empty:
            return []
        i0, i1 = self._domain_indexes(domain)
        return self._points[i0:i1]

    def _did_change_points(self):
        self.interval = None
        len_points = len(self._points)
        if len_points > 1:
            self.interval = (self._points[-1][0] - self._points[0][0]) / (len_points - 1)

    def _points_equally_spaced(self, old_points, new_points):
        old_points_len = len(old_points)
        new_points_len = len(new_points)
        if new_points_len == 0 or old_points_len + new_points_len < 2:
            return True

        if old_points_len != 0:
            start = old_points[-1][0]
            steps = new_points_len
        else:
            start = new_points[0][0]
            steps = new_points_len - 1

        if self.interval is not None:
            interval = self.interval
        elif old_points_len == 1:
            interval = new_points[0][0] - old_points[-1][0]
        else:
            interval = new_points[1][0] - new_points[0][0]

        expected_last_x = start + interval * steps
        return new_points[-1][0] == expected_last_x

    def y(self, x):
        if not self.domain.contains(x):
            return None
        i_ = self.x_index(x)
        u = i_ % 1.0
        i = int(i_)
        if u == 0:
            return self._points[i][1]
        if self.interpolation == Points.interpolation.linear:
            y1 = self._points[i][1]
            y2 = self._points[i + 1][1]
            if y1 is None or y2 is None:
                return None
            return (1.0 - u) * y1 + u * y2
        elif self.interpolation == Points.interpolation.previous:
            return self._points[i][1]
        elif self.interpolation == Points.interpolation.next:
            return self._points[i + 1][1]
        else:
            raise Exception('Unknown interpolation')

    def y_start(self):
        if self.domain.is_empty:
            return None
        return self._points[0][1]

    def y_end(self):
        if self.domain.is_empty:
            return None
        return self._points[-1][1]

    # TODO: optimise d_y()

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        if self.domain.is_empty:
            return None
        min_step = self.resolve_min_step(min_step)
        if not self.domain.contains(x + min_step, enforce_start=False):
            return None
        if math.isinf(x) and x < 0:
            return self.domain.start
        # i = max(0, int(math.floor(self.x_index(x))) + 1)
        i = max(0, int(math.ceil(self.x_index(x + min_step))))
        return self._points[i][0]

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        if self.domain.is_empty:
            return None
        min_step = self.resolve_min_step(min_step)
        if not self.domain.contains(x - min_step, enforce_end=False):
            return None
        if math.isinf(x) and x > 0:
            return self.domain.end
        # i = min(len(self._points) - 1, int(math.ceil(self.x_index(x))) - 1)
        i = min(len(self._points) - 1, int(math.floor(self.x_index(x - min_step))))
        return self._points[i][0]

    def x_index(self, x):
        if self.is_uniform:
            if self.interval is None:
                return 0
            return (x - self._points[0][0]) / self.interval
        else:
            i1 = _bisect_points(self._points, x)
            i0 = i1 - 1
            if i0 < 0:
                return i1
            if i1 >= len(self._points):
                return i0
            x0 = self._points[i0][0]
            x1 = self._points[i1][0]
            u = (x - x0) / (x1 - x0)
            return float(i0) + u

    def _domain_indexes(self, domain):
        """
        Turn a domain into start and end indexes (inclusive and exclusive respectively).
        """
        points_len = len(self._points)
        if domain.is_superset_of(self.domain):
            return 0, points_len

        domain = Interval.intersection([domain, self.domain])
        if domain.is_empty:
            return 0, 0

        if domain.is_negative_infinite:
            start_i = 0
        else:
            i = max(0, int(math.floor(self.x_index(domain.start))))
            while i < points_len:
                x = self._points[i][0]
                if x >= domain.end:
                    # i -= 1
                    break
                if domain.contains(x):
                    break
                i += 1
            start_i = i
            
        if domain.is_positive_infinite:
            end_i = points_len
        else:
            i = min(points_len, int(math.ceil(self.x_index(domain.end))))
            while i >= 0:
                x = self._points[i][0]
                if x <= domain.start:
                    # i += 1
                    break
                if domain.contains(x):
                    break
                i -= 1
            end_i = i + 1

        return start_i, end_i

def _bisect_points(a, x, lo=0, hi=None):
    """
    Insert point `p` in list `a`, and keep it sorted assuming `a` is sorted.
    If `p` is already in `a`, insert it to the left of the leftmost `p` if `no_repeat` is `False`.
    Optional args `lo` (default `0`) and `hi` (default `len(a)`) bound the
    slice of `a` to be searched.

    Source: https://github.com/python/cpython/blob/master/Lib/bisect.py
    """

    if lo < 0:
        raise ValueError('lo must be non-negative')
    a_len = len(a)
    if hi is None:
        hi = a_len
    while lo < hi:
        mid = (lo + hi) // 2
        if a[mid][0] < x: lo = mid + 1
        else: hi = mid
    return lo



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['curvepy', 'curvepy.extension']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.12.1,<0.13.0',
 'intervalpy>=0.1.0,<0.2.0',
 'numpy>=1.15.0,<2.0.0',
 'pyduration>=0.1.0,<0.2.0']

entry_points = \
{'console_scripts': ['test = pytest:main']}

setup_kwargs = {
    'name': 'curvepy',
    'version': '0.1.0',
    'description': 'A mathematical and financial function curve utility library.',
    'long_description': '# curvepy\n\n![Tests](https://github.com/diatche/curvepy/workflows/Tests/badge.svg)\n\nA mathematical and financial function curve utility library for Python.\n\n# Installation\n\nWith [poetry](https://python-poetry.org):\n\n```bash\npoetry add curvepy\n```\n\nOr with pip:\n\n```\npip3 install curvepy\n```\n\n# Usage\n\nHave a look at the [documentation](https://diatche.github.io/curvepy/).\n\nBasic usage:\n\n```python\n# Create a line\nfrom curvepy import Line\n\nline = Line(const=1, slope=2)\nassert line.y(0) == 1\nassert line.y(1) == 3\n\n# Function arithmetic\nline2 = Line(const=-1, slope=-2)\nline_sum = line1 + line2\nassert line_sum.y(0) == 0\nassert line_sum.y(1) == 0\n```\n\n# Development\n\n## Updating Documentation\n\nThe module [pdoc3](https://pdoc3.github.io/pdoc/) is used to automatically generate documentation. To update the documentation:\n\n1. Install `pdoc3` if needed with `pip3 install pdoc3`.\n2. Navigate to project root and install dependencies: `poetry install`.\n3. Generate documetation files with: `pdoc3 -o docs --html curvepy`.\n4. The new files will be in `docs/curvepy`. Move them to `docs/` and replace existing files.\n',
    'author': 'Pavel Diatchenko',
    'author_email': 'diatche@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/diatche/func',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4',
}


setup(**setup_kwargs)



from .curve import Curve, MIN_STEP
from intervalpy import Interval

class Map(Curve):

    def get_domain(self):
        return self.curve.domain

    def __init__(self, func, tfm, skip_none=False, min_step=MIN_STEP, name=None):
        if bool(func.min_step) and func.min_step > min_step:
            min_step = func.min_step
        super().__init__(min_step=min_step)
        self.curve = Curve.parse(func)
        self.skip_none = skip_none
        self.name = name

        tfm_ags = type(self).count_positional_args(tfm)
        if tfm_ags == 0 or tfm_ags > 2:
            raise Exception('Unable to adapt function')
        self._map_tfm_with_x = tfm_ags > 1
        self.map_tfm = tfm

        self._observer_token = self.curve.add_observer(begin=self.begin_update, end=self.end_update, prioritize=True)
    
    def __repr__(self):
        try:
            name = self.name or f'map({self.map_tfm.__name__})'
            return f'{self.curve}.{name}'
        except Exception as e:
            return super().__repr__() + f'({e})'

    def __del__(self):
        self.curve.remove_observer(self._observer_token)

    def y(self, x):
        y = self.curve.y(x)
        if y is None and self.skip_none:
            return None
        return self._map(x, y)

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.curve.x_previous(x, min_step=min_step, limit=limit)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.curve.x_next(x, min_step=min_step, limit=limit)

    def _map(self, x, y):
        if self._map_tfm_with_x:
            return self.map_tfm(x, y)
        else:
            return self.map_tfm(y)


from .extension import Extension
from intervalpy import Interval
from ..const import GOLD

class WeaveExtension(Extension):

    """
    Extends an end of a function with a line using its edge tangents.
    """

    name = "weave"

    def __init__(self, func, period=None, degree=3, tolerance=1e-3, **kwargs):
        super().__init__(func, **kwargs)
        if self.start:
            raise Exception('Weave extension currently only supports extending the end')
        self.period = float(period)
        self.degree = int(degree)
        self.tolerance = float(tolerance)
        assert self.period > 0
        assert self.degree > 1
        assert self.tolerance > 0
        if self.end:
            self.setup()

    def y_extension(self, x):
        if self.end and x >= self.curve.domain.end:
            self.accumulate(x)
            return self.extension.y(x)
        return None

    def update_extension(self):
        if self.end:
            self.extension.reset()
            self.extension_interval = self.get_extension_interval()

    def get_extension_interval(self):
        if self.end and not self.curve.domain.is_empty:
            smas = list(map(lambda sma: sma.y(self.curve.domain.end), self.smas))
            if len(list(filter(None, smas))) == len(smas):
                return Interval.gte(self.curve.domain.end)
        # Not enough data
        return Interval.empty()

    def accumulate(self, x):
        x0 = self.layers.domain.end
        while True:
            if x0 is None or x0 >= x:
                break
            x1 = self.accumulate_next(x0)
            assert x1 > x0
            x0 = x1

    def setup(self):
        self.periods = []
        self.sma_tangents = []
        self.smas = []
        self.macs = []
        mac_period_coef = 2.5
        for i in range(self.price_degree):
            period = self.quote_period * GOLD ** float(i * 2)
            period = round(period / self.quote_period) * self.quote_period
            self.periods.insert(0, period)

        for i in range(self.price_degree):
            period = self.periods[i]
            sma = self.market.quote_func.ohlc.sma(period, is_period=True)
            sma = sma.offset(period * -0.5)
            sma_tangent = sma.extension('tangent', start=False, end=True)
            self.sma_tangents.append(sma_tangent)
            if i == 0:
                sma = sma_tangent
            if i != 0:
                mac = (sma - self.sma_tangents[i - 1]).extension('sin', period=period * mac_period_coef, start=False, end=True)
                sma = self.sma_tangents[i - 1] + mac
                self.smas.append(sma)
                self.macs.append(mac)
            self.smas.append(sma)

    def accumulate_next(self, x):
        from scipy.optimize import minimize
        
        x0 = self.layers.x_previous(x, min_step=self.min_step)
        if x0 is None:
            return None
        y = self.layers.y(x)
        step = x - x0
        x1 = x + step

        def get_values():
            # return list(map(lambda mac: (mac.y(x1) - mac.y(x)) / step, self.macs))
            # return list(map(lambda mac: mac.y(x1), self.macs)) + list(map(lambda sma: sma.y(x1), self.smas))
            # return list(map(lambda sma: sma.y(x1), self.smas))
            return list(map(lambda mac: mac.y(x1), self.macs))
            # return list(map(lambda mac: (mac.y(x1) - mac.y(x)) / step, self.macs))

        start_values = get_values()
        start_values = list(map(lambda y: y * 0.9, start_values))
        values_len = len(start_values)

        smas = list(map(lambda sma: sma.y(x1), self.smas))
        sma_len = len(smas)

        def error_func(params):
            y_coef = params[0]
            y1 = y * (1 + y_coef)
            self.extension.replace((x1, y1), or_append=True)

            error_sum = 0

            values = get_values()
            for i in range(values_len):
                error_sum += (start_values[i] - values[i]) ** 2
                # error_sum += (values[i]) ** 2

            # for i in range(sma_len):
            #     error_sum += (y1 - smas[i]) ** 2
            #     # error_sum += (values[i]) ** 2

            return error_sum

        # docs: https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize
        params = [0]
        method = 'Nelder-Mead'
        result = minimize(error_func, params, method=method, tol=self.tolerance)
        
        y_coef = result.x[0]
        y1 = y * (1 + y_coef)
        self.extension.replace((x1, y1), or_append=True)

        return x1

    # def setup(self):
    #     self.extension = Points([])
    #     self.layers = Curve.first([self.curve, self.extension])
    #     self.periods = []
    #     self.smas = []
    #     self.macs = []
    #     mac_period_coef = 2.5
    #     for i in range(self.degree):
    #         period = self.period * GOLD ** float(i * 2)
    #         period = round(period / self.period) * self.period
    #         self.periods.insert(0, period)

    #     for i in range(self.degree):
    #         period = self.periods[i]
    #         sma = self.layers.sma(period, is_period=True)
    #         sma = sma.offset(period * -0.5)
    #         if i == 0:
    #             sma = sma.extension('tangent', start=False, end=True)
    #         else:
    #             mac = (sma - self.smas[i - 1]).extension('sin', period=period * mac_period_coef, start=False, end=True)
    #             sma = self.smas[i - 1] + mac
    #             self.macs.append(mac)
    #         self.smas.append(sma)

    # def accumulate_next(self, x):
    #     x0 = self.layers.x_previous(x, min_step=self.min_step)
    #     if x0 is None:
    #         return None
    #     y = self.layers.y(x)
    #     step = x - x0
    #     x1 = x + step

    #     def get_values():
    #         # return list(map(lambda mac: (mac.y(x1) - mac.y(x)) / step, self.macs))
    #         # return list(map(lambda mac: mac.y(x1), self.macs)) + list(map(lambda sma: sma.y(x1), self.smas))
    #         # return list(map(lambda sma: sma.y(x1), self.smas))
    #         return list(map(lambda mac: mac.y(x1), self.macs))
    #         # return list(map(lambda mac: (mac.y(x1) - mac.y(x)) / step, self.macs))

    #     start_values = get_values()
    #     start_values = list(map(lambda y: y * 0.9, start_values))
    #     values_len = len(start_values)

    #     smas = list(map(lambda sma: sma.y(x1), self.smas))
    #     sma_len = len(smas)

    #     def error_func(params):
    #         y_coef = params[0]
    #         y1 = y * (1 + y_coef)
    #         self.extension.replace((x1, y1), or_append=True)

    #         error_sum = 0

    #         values = get_values()
    #         for i in range(values_len):
    #             error_sum += (start_values[i] - values[i]) ** 2
    #             # error_sum += (values[i]) ** 2

    #         # for i in range(sma_len):
    #         #     error_sum += (y1 - smas[i]) ** 2
    #         #     # error_sum += (values[i]) ** 2

    #         return error_sum

    #     # docs: https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize
    #     params = [0]
    #     method = 'Nelder-Mead'
    #     result = minimize(error_func, params, method=method, tol=self.tolerance)
        
    #     y_coef = result.x[0]
    #     y1 = y * (1 + y_coef)
    #     self.extension.replace((x1, y1), or_append=True)

    #     return x1



import math
from .scan import Scan
from .points import Points
from .curve import MIN_STEP
from intervalpy import Interval

class Accumulator(Scan):

    def __init__(
            self,
            func,
            tfm,
            interpolation=None,
            uniform=True,
            min_step=MIN_STEP):
        assert callable(tfm)
        self.accumulated_points = Points([], interpolation=interpolation, uniform=uniform)
        self.accumulator_transform = tfm
        super().__init__(func, self._accumulate, min_step=min_step)

    def scanned_y(self, x):
        return self.accumulated_points.y(x)

    def reset_scan(self):
        super().reset_scan()
        self.accumulated_points.reset()

    def x_previous(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.accumulated_points.x_previous(x, min_step=min_step, limit=limit) or self.curve.x_previous(x, min_step=min_step, limit=limit)

    def x_next(self, x, min_step=MIN_STEP, limit=None):
        min_step = self.resolve_min_step(min_step)
        return self.accumulated_points.x_next(x, min_step=min_step, limit=limit) or self.curve.x_next(x, min_step=min_step, limit=limit)

    def begin_update(self, domain):
        self.accumulated_points.reset(domain)
        super().begin_update(domain)

    def _accumulate(self, x, y):
        if self.accumulated_points.domain.is_empty:
            last_y = None
        else:
            last_x = self.accumulated_points.domain.end
            assert last_x < x
            last_y = self.accumulated_points.y(last_x)
        new_y = self.accumulator_transform(x, y, last_y)
        self.accumulated_points.append((x, new_y))
